#include "global.h"
#include "cont.h"
#include "input.h"
#include "settings.h"
#include "netplay.h"
#include "paste.h"
#include "state.h"
#include "reverse.h"
#include "draw.h"
#include "resource.h"
#include "main.h"
#include "msx.h"
#include "movie.h"
#include "z80.h"


/* controller names */
static const char* cont_name[CONT_D_MAX]={
	"joystick",
	"Taito Arkanoid pad",
	"Konami HyperShot",
	"Sony Magic Key",
	
	"keyboard",
	
	"nothing"
};
const char* cont_get_name(u32 i) { if (i>=CONT_D_MAX) return NULL; else return cont_name[i]; }

static const char* cont_a_name[CONT_DA_MAX]={
	"joystick (port 1)",
	"joystick (port 2)",
	"Taito Arkanoid pad (port 1)",
	"Taito Arkanoid pad (port 2)",
	"Konami HyperShot controller (port 1)",
	"Konami HyperShot controller (port 2)",
	
	"keyboard"
};
const char* cont_get_a_name(u32 i) { if (i>=CONT_DA_MAX) return NULL; else return cont_a_name[i]; }

/* keyboard regions */
static int cont_region=CONT_REGION_INTERNATIONAL;
static const char* cont_region_name[CONT_REGION_MAX]={
	"International",	"Japanese",	"Korean",	"United Kingdom"
};
static const char* cont_region_shortname[CONT_REGION_MAX]={
	"I",			"JP",		"KR",		"UK"
};
const char* cont_get_region_name(u32 i) { if (i>=CONT_REGION_MAX) return NULL; else return cont_region_name[i]; }
const char* cont_get_region_shortname(u32 i) { if (i>=CONT_REGION_MAX) return NULL; else return cont_region_shortname[i]; }
int cont_get_region(void) { return cont_region; }
void cont_set_region(u32 r) { if (r>=CONT_REGION_MAX) r=0; cont_region=r; }

/* ports */
static const int cont_port_default[CONT_P_MAX]={CONT_D_J,CONT_D_J,CONT_D_KEY};
static int cont_port[CONT_P_MAX];
u32 cont_get_port(u32 p) { if (p<CONT_P_MAX) return cont_port[p]; else return CONT_D_NOTHING; }

/* trigger info */
static const char* cont_t_info[CONT_T_MAX]={
	/* joystick 1 */
	"joystick 1 Up",
	"joystick 1 Down",
	"joystick 1 Left",
	"joystick 1 Right",
	"joystick 1 Button A",
	"joystick 1 Button B",
	
	/* joystick 2 */
	"joystick 2 Up",
	"joystick 2 Down",
	"joystick 2 Left",
	"joystick 2 Right",
	"joystick 2 Button A",
	"joystick 2 Button B",
	
	/* hypershot 1 */
	"HyperShot 1 Jump",
	"HyperShot 1 Run",
	
	/* hypershot 2 */
	"HyperShot 2 Jump",
	"HyperShot 2 Run",
	
	/* arkanoid */
	"Arkanoid pad 1 Button",
	"Arkanoid pad 2 Button",
	
	/* keyboard */
	"keyboard 0",
	"keyboard 1",
	"keyboard 2",
	"keyboard 3",
	"keyboard 4",
	"keyboard 5",
	"keyboard 6",
	"keyboard 7",
	"keyboard 8",
	"keyboard 9",
	
	"keyboard A ",
	"keyboard B ",
	"keyboard C ",
	"keyboard D ",
	"keyboard E ",
	"keyboard F ",
	"keyboard G ",
	"keyboard H ",
	"keyboard I ",
	"keyboard J ",
	"keyboard K ",
	"keyboard L ",
	"keyboard M ",
	"keyboard N ",
	"keyboard O ",
	"keyboard P ",
	"keyboard Q ",
	"keyboard R ",
	"keyboard S ",
	"keyboard T ",
	"keyboard U ",
	"keyboard V ",
	"keyboard W ",
	"keyboard X ",
	"keyboard Y ",
	"keyboard Z ",
	
	"keyboard F1",
	"keyboard F2",
	"keyboard F3",
	"keyboard F4",
	"keyboard F5",
	
	"keyboard `",
	"keyboard -",
	"keyboard =",
	"keyboard [",
	"keyboard ]",
	"keyboard \\",
	"keyboard ;",
	"keyboard '",
	"keyboard ,",
	"keyboard .",
	"keyboard /",
	"keyboard Accent Mark/_",
	
	"keyboard Spacebar",
	"keyboard Tab",
	"keyboard Return",
	"keyboard Home",
	"keyboard Insert",
	"keyboard Delete",
	"keyboard Backspace",
	"keyboard Escape",
	"keyboard Stop",
	"keyboard Select",
	
	"keyboard Up",
	"keyboard Down",
	"keyboard Left",
	"keyboard Right",
	
	"keyboard Shift",
	"keyboard Caps Lock",
	"keyboard Control",
	"keyboard Graph",
	"keyboard Code/Kana Lock",
	
	"keyboard Numpad 0",
	"keyboard Numpad 1",
	"keyboard Numpad 2",
	"keyboard Numpad 3",
	"keyboard Numpad 4",
	"keyboard Numpad 5",
	"keyboard Numpad 6",
	"keyboard Numpad 7",
	"keyboard Numpad 8",
	"keyboard Numpad 9",
	"keyboard Numpad ,",
	"keyboard Numpad .",
	"keyboard Numpad +",
	"keyboard Numpad -",
	"keyboard Numpad *",
	"keyboard Numpad /",
	
	"keyboard Torikeshi",
	"keyboard Jikkou",
	
	/* extra keyboard keys */
	"keyboard Sony Pause"
};
const char* cont_get_trigger_info(u32 i) { if (i>=CONT_T_MAX) return input_get_trigger_info(i-CONT_T_MAX); else return cont_t_info[i]; }

/* axis info */
static const char* cont_a_info[CONT_A_MAX]={
	"Arkanoid pad 1 Dial",
	"Arkanoid pad 2 Dial"
};
const char* cont_get_axis_info(u32 i) { if (i>=CONT_A_MAX) return input_get_axis_info(i-CONT_A_MAX); else return cont_a_info[i]; }


u32 cont_get_da(u32 p)
{
	u32 ret=(CONT_DA_EMPTY)|(CONT_DA_EMPTY<<8)|(CONT_DA_EMPTY<<16)|(CONT_DA_EMPTY<<24);
	
	/* no need to add user-inputless devices like magic key */
	switch (p) {
		
		/* port 1 */
		case CONT_P1:
			switch (cont_get_port(CONT_P1)) {
				case CONT_D_J: ret=(ret&0xffffff00)|CONT_DA_J1; break; /* standard joystick */
				case CONT_D_HS: ret=(ret&0xffffff00)|CONT_DA_HS1; break; /* hypershot */
				case CONT_D_ARK: ret=(ret&0xffffff00)|CONT_DA_ARK1; break; /* arkanoid */
				default: break;
			}
			break;
		
		/* port 2 */
		case CONT_P2:
			switch (cont_get_port(CONT_P2)) {
				case CONT_D_J: ret=(ret&0xffffff00)|CONT_DA_J2; break; /* standard joystick */
				case CONT_D_HS: ret=(ret&0xffffff00)|CONT_DA_HS2; break; /* hypershot */
				case CONT_D_ARK: ret=(ret&0xffffff00)|CONT_DA_ARK2; break; /* arkanoid */
				default: break;
			}
			break;
		
		/* keyboard */
		case CONT_PKEY:
			switch (cont_get_port(CONT_PKEY)) {
				case CONT_D_KEY: ret=(ret&0xffffff00)|CONT_DA_KEY; break; /* keyboard */
				default: break;
			}
			break;
		
		default: break;
	}
	
	return ret;
}


void cont_init(void)
{
	char* c=NULL;
	int r,port;
	
	/* keyboard region */
	r=CONT_REGION_INTERNATIONAL;
	SETTINGS_GET_STRING(settings_info(SETTINGS_REGION),&c);
	if (c&&strlen(c)) {
		for (r=0;r<CONT_REGION_MAX;r++) if (stricmp(c,cont_get_region_name(r))==0) break;
		if (r==CONT_REGION_MAX) r=CONT_REGION_INTERNATIONAL;
	}
	MEM_CLEAN(c);
	cont_region=r;
	
	/* get inserted controllers settings (no custom keyboard yet) */
	cont_port[CONT_PKEY]=CONT_D_KEY;
	
	for (port=CONT_P1;port<(CONT_P_MAX-1);port++) {
		int setport=0;
		
		switch (port) {
			case CONT_P1: setport=SETTINGS_PORT1; break;
			case CONT_P2: setport=SETTINGS_PORT2; break;
			default: break;
		}
		
		c=NULL;
		SETTINGS_GET_STRING(settings_info(setport),&c);
		cont_port[port]=cont_port_default[port];
		
		if (c&&strlen(c)) {
			int cont;
			for (cont=0;cont<CONT_D_MAX;cont++) if (stricmp(c,cont_name[cont])==0) break;
			if (cont==CONT_D_KEY||cont==CONT_D_MAX) cont=cont_port_default[port];
			cont_port[port]=cont;
		}
		MEM_CLEAN(c);
		
		cont_insert(port,cont_port[port],TRUE,TRUE);
	}
}

void cont_clean(void)
{
	;
}

void cont_settings_save(void)
{
	/* called from input.c input_settings_save */
	SETTINGS_PUT_STRING(NULL,"\n; Valid port options: ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_J)); SETTINGS_PUT_STRING(NULL,", ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_HS)); SETTINGS_PUT_STRING(NULL,", ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_MAGIC)); SETTINGS_PUT_STRING(NULL,",\n; -- ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_ARK)); SETTINGS_PUT_STRING(NULL,", or ");
	SETTINGS_PUT_STRING(NULL,cont_get_name(CONT_D_NOTHING)); SETTINGS_PUT_STRING(NULL,".\n\n");
	
	SETTINGS_PUT_STRING(settings_info(SETTINGS_PORT1),cont_name[cont_port[CONT_P1]]);
	SETTINGS_PUT_STRING(settings_info(SETTINGS_PORT2),cont_name[cont_port[CONT_P2]]);
	
	SETTINGS_PUT_STRING(NULL,"\n; Valid "); SETTINGS_PUT_STRING(NULL,settings_info(SETTINGS_REGION)); SETTINGS_PUT_STRING(NULL," options: ");
	SETTINGS_PUT_STRING(NULL,cont_get_region_name(CONT_REGION_INTERNATIONAL)); SETTINGS_PUT_STRING(NULL,", ");
	SETTINGS_PUT_STRING(NULL,cont_get_region_name(CONT_REGION_JAPANESE)); SETTINGS_PUT_STRING(NULL,", ");
	SETTINGS_PUT_STRING(NULL,cont_get_region_name(CONT_REGION_KOREAN)); SETTINGS_PUT_STRING(NULL,",\n; -- or ");
	SETTINGS_PUT_STRING(NULL,cont_get_region_name(CONT_REGION_UK)); SETTINGS_PUT_STRING(NULL,".\n\n");
	
	SETTINGS_PUT_STRING(settings_info(SETTINGS_REGION),cont_get_region_name(cont_region));
}


/* emu */
static u8 keyrow[16];
static int keyextra;		/* extra keys not on matrix, like sony pause key */
static int stateextra;		/* extra state, like sony pause state */
static int sonypause_prev;

static int curport;
static int gen_output[2];
static u8 joy_state[2];
static u8 hs_state[2];

#define ARK_DIAL_MIN 163
#define ARK_DIAL_MAX 309
static int ark_shift_upd[2];
static int ark_shift[2];
static int ark_button[2];
static int ark_bitcount[2];

u8 __fastcall cont_get_joystate(u8 j) { return joy_state[j]; }
u8 __fastcall cont_get_keyrow(u8 row) { return keyrow[row]; }
int __fastcall cont_get_keyextra(void) { return keyextra; }
int __fastcall cont_get_stateextra(void) { return stateextra; }
void __fastcall cont_set_keyextra(int e) { keyextra=e; }
void __fastcall cont_set_stateextra(int e){ stateextra=e; }

int __fastcall cont_check_keypress(void)
{
	/* ret:
		bit 0:	read trigger pressed (has priority over bit 1)
		bit 1:	raw axis moved
	*/
	
	/* keyboard */
	int i=12;
	while (i--) if (keyrow[i]!=0xff) return 1;
	if (keyextra) return 1;
	
	/* ports */
	i=2;
	while (i--) {
		switch (cont_port[i]) {
			
			/* joystick */
			case CONT_D_J:
				if (joy_state[i]!=0x3f) return 1;
				break;
			
			/* hypershot */
			case CONT_D_HS:
				if (hs_state[i]!=0x3f) return 1;
				break;
			
			/* arkanoid */
			case CONT_D_ARK: {
				int ret=((input_get_axis(CONT_A_ARK1_DIAL+i*CONT_A_ARK_SIZE))!=0)<<1|(ark_button[i]>>1&1);
				if (ret&1) return 1;
				else if (ret) return ret;
				break;
			}
			
			default: break;
		}
	}
	
	return 0;
}

/* POWERON */
static void cont_poweron_port(int port)
{
	/* joystick */
	joy_state[port]=0x3f;
	
	/* hypershot */
	hs_state[port]=0x3f;
	
	/* arkanoid */
	ark_shift_upd[port]=0xec;
	ark_shift[port]=0x1d9;
	ark_button[port]=2;
	ark_bitcount[port]=9;
}

void cont_poweron(void)
{
	curport=gen_output[0]=gen_output[1]=0;
	
	memset(keyrow,0xff,16);
	keyextra=stateextra=0;
	cont_reset_sonypause();
	cont_poweron_port(0); cont_poweron_port(1);
}

/* UPDATE */
typedef void __fastcall (*_contupdate)(int);
static _contupdate contupdate[2];
static void __fastcall contupdate_nothing(int port) { ; } /* nothing */

/* joystick: 4 directions, 1 or 2 buttons (many different models) */
static void __fastcall contupdate_joystick(int port)
{
	const int size=port*CONT_T_J_SIZE;
	joy_state[port]=(input_trigger_held(CONT_T_J1_UP+size)|(input_trigger_held(CONT_T_J1_DOWN+size)<<1)|(input_trigger_held(CONT_T_J1_LEFT+size)<<2)|(input_trigger_held(CONT_T_J1_RIGHT+size)<<3)|(input_trigger_held(CONT_T_J1_A+size)<<4)|(input_trigger_held(CONT_T_J1_B+size)<<5))^0x3f;
	if (draw_is_flip_x()) joy_state[port]=(joy_state[port]&0xf3)|(joy_state[port]<<1&8)|(joy_state[port]>>1&4);
}

/* hypershot: jump button on the left, run button on the right (device can be turned around to 'swap' buttons) */
static void __fastcall contupdate_hypershot(int port)
{
	const int size=port*CONT_T_HS_SIZE;
	hs_state[port]=(input_trigger_held(CONT_T_HS1_JUMP+size)<<4^0x30)|(input_trigger_held(CONT_T_HS1_RUN+size)?5:0xf); /* same as joystick button 1 and right+down */
}

/* arkanoid: dial on the bottom, button above (only usable on ROM release, CAS release works with joystick) */
static void __fastcall contupdate_arkanoid(int port)
{
	const int tsize=port*CONT_T_ARK_SIZE;
	const int asize=port*CONT_A_ARK_SIZE;
	long l=input_get_axis(CONT_A_ARK1_DIAL+asize);
	
	if (draw_is_flip_x()) l=-l; /* switch left/right done in advance */
	l+=ark_shift_upd[port];
	CLAMP(l,ARK_DIAL_MIN,ARK_DIAL_MAX); /* 163: clear level on left? (untested), 164: far left, 236: middle, 308: far right, 309: clear level on right */
	
	ark_shift_upd[port]=l;
	ark_button[port]=input_trigger_held(CONT_T_ARK1_BUTTON+tsize)<<1;
}

/* general */
void __fastcall cont_update(void)
{
	int i=11; /* matrix */
	
	/* update plugged-in devices */
	contupdate[0](0); contupdate[1](1);
	
	/* keyboard.. the matrix reloaded ;p */
	#define k(row,k7,k6,k5,k4,k3,k2,k1,k0) keyrow[row]=~(input_trigger_held(CONT_T_K_##k7)<<7|input_trigger_held(CONT_T_K_##k6)<<6|input_trigger_held(CONT_T_K_##k5)<<5|input_trigger_held(CONT_T_K_##k4)<<4|input_trigger_held(CONT_T_K_##k3)<<3|input_trigger_held(CONT_T_K_##k2)<<2|input_trigger_held(CONT_T_K_##k1)<<1|input_trigger_held(CONT_T_K_##k0))
	k(0,  7,         6,        5,         4,         3,      2,      1,     0);
	k(1,  SEMICOLON, RBRACKET, LBRACKET,  BACKSLASH, EQUALS, MINUS,  9,     8);
	k(2,  B,         A,        ACCENT,    SLASH,     PERIOD, COMMA,  TILDE, QUOTE);
	k(3,  J,         I,        H,         G,         F,      E,      D,     C);
	k(4,  R,         Q,        P,         O,         N,      M,      L,     K);
	k(5,  Z,         Y,        X,         W,         V,      U,      T,     S);
	k(6,  F3,        F2,       F1,        CODE,      CAPS,   GRAPH,  CTRL,  SHIFT);
	k(7,  RETURN,    SELECT,   BACKSPACE, STOP,      TAB,    ESC,    F5,    F4);
	k(8,  RIGHT,     DOWN,     UP,        LEFT,      DEL,    INS,    HOME,  SPACE);
	k(9,  N4,        N3,       N2,        N1,        N0,     NSLASH, NPLUS, NASTERISK);
	k(10, NPERIOD,   NCOMMA,   NMINUS,    N9,        N8,     N7,     N6,    N5);
	#undef k
	keyrow[11]=~((input_trigger_held(CONT_T_K_TORIKESHI)<<3)|(input_trigger_held(CONT_T_K_JIKKOU)<<1)); /* row 12, only 2 keys */
	
	/* extra keys, only sony pause for now */
	keyextra=input_trigger_held(CONT_T_K_SONYPAUSE)&CONT_EKEY_SONYPAUSE;
	
	/* switch keyboard left/right */
	if (draw_is_flip_x()) keyrow[8]=(keyrow[8]&0x6f)|(keyrow[8]<<3&0x80)|(keyrow[8]>>3&0x10);
	
	reverse_cont_update();
	paste_frame(keyrow);
	netplay_frame(keyrow,&keyextra,joy_state);
	
	/* matrix quirk (after netplay) */
	while (i--) {
		
		if (keyrow[i]==0xff) continue;
		
		if (i==6) {
			/* modifier keys */
			int k=keyrow[6]&0x15; keyrow[6]|=0x15;
			if (keyrow[6]!=0xff) {
				int j=12;
				while (j--) {
					if ((keyrow[6]|keyrow[j])!=0xff) keyrow[6]=keyrow[j]=keyrow[6]&keyrow[j];
				}
			}
			keyrow[6]=k|(keyrow[6]&0xea);
		}
		
		else {
			int j=12;
			while (j--) {
				if ((keyrow[i]|keyrow[j])!=0xff) keyrow[i]=keyrow[j]=keyrow[i]&keyrow[j];
			}
		}
	}
	
	movie_frame();
}

/* READ/WRITE */
typedef u8 __fastcall (*_contread)(int);
static _contread contread[2];
u8 __fastcall cont_read(void)
{
	/* PSG R14 bit 0-5 read, pins 6/7 input ANDed with pins 6/7 output */
	return contread[curport](curport)&((gen_output[curport]<<4&0x30)|0xf);
}

typedef void __fastcall (*_contwrite)(int,u8);
static _contwrite contwrite[2];
void __fastcall cont_write(u8 v)
{
	/* PSG R15 write, output pins 6,7,8 */
	int v0=(v&3)|(v>>2&4);
	int v1=(v>>2&3)|(v>>3&4);
	
	curport=v>>6&1;
	
	contwrite[0](0,v0);
	contwrite[1](1,v1);
	
	gen_output[0]=v0;
	gen_output[1]=v1;
}

/* nothing */
static u8 __fastcall contread_nothing(int port) { return 0x3f; }
static void __fastcall contwrite_nothing(int port,u8 v) { ; }

/* joystick
 Some MSX joysticks/pads have a similar quirk to the hypershot controller, where pins 6/8 influence direction pins.
 Since there are so many joystick models, it's left unemulated */
static u8 __fastcall contread_joystick(int port) { return joy_state[port]; }

/* hypershot */
static u8 __fastcall contread_hypershot(int port)
{
	/* if output pin 8 is high and output pin 6 is low, the run button will only be detected if the jump button is pressed */
	if (gen_output[port]&4&&!(gen_output[port]&1)&&hs_state[port]&0x10) return 0x3f;
	else return hs_state[port];
}

/* magic key (simple cheat device by Sony, not many MSX1 games support it, the only known one being Gall Force) */
static u8 __fastcall contread_magickey(int port) { return 0x3c; }

/* arkanoid */
static u8 __fastcall contread_arkanoid(int port)
{
	/* bit 0: dial position (serial read), bit 1: button */
	return (ark_shift[port]>>ark_bitcount[port]&1)|ark_button[port]|0x3c;
}
static void __fastcall contwrite_arkanoid(int port,u8 v)
{
	/* rising edge pin 6 */
	if (v&1&~gen_output[port]) ark_bitcount[port]-=(ark_bitcount[port]!=0);
	
	/* rising edge pin 8 */
	if (v&4&~gen_output[port]) {
		ark_bitcount[port]=9;
		ark_shift[port]=ark_shift_upd[port]<<1|1;
	}
}

/* MISC */
/* sony pause state */
void cont_mute_sonypause(signed short* stream,int len)
{
	if (!(stateextra&CONT_ESTATE_SONYPAUSE)) return;
	
	/* mute internal audio signal, eg. shouldn't mute moonsound */
	while (len--) stream[len]=0;
}
void cont_set_sonypause(int p)
{
	z80_set_busrq(p); /* halt/unhalt z80 */
	stateextra=(stateextra&~CONT_ESTATE_SONYPAUSE)|p;
}
void cont_reset_sonypause(void)
{
	sonypause_prev=0xff;
	cont_set_sonypause(FALSE);
}
void cont_check_sonypause(void)
{
	int prev=(keyextra&CONT_EKEY_SONYPAUSE)|(stateextra&CONT_ESTATE_SONYPAUSE)<<1;
	if (~sonypause_prev&keyextra&CONT_EKEY_SONYPAUSE) cont_set_sonypause(~sonypause_prev>>1&1); /* rising edge */
	sonypause_prev=prev;
}

/* INSERT */
void cont_insert(int port,int d,int init,int poweron)
{
	UINT id=IDM_PORT_FIRST+port*IDM_PORT_SIZE;
	
	if (!init&&cont_port[port]==d) return;
	cont_port[port]=d;
	if (poweron) cont_poweron_port(port);
	
	/* menu:
		Joystick		+0
		Konami HyperShot	+1
		Sony Magic Key		+2
		Taito Arkanoid Pad	+3
		Nothing			+4
	*/
	
	switch (d) {
		case CONT_D_J:
			id+=0;
			contupdate[port]=contupdate_joystick;
			contread[port]=contread_joystick;
			contwrite[port]=contwrite_nothing;
			break;
		case CONT_D_HS:
			id+=1;
			contupdate[port]=contupdate_hypershot;
			contread[port]=contread_hypershot;
			contwrite[port]=contwrite_nothing;
			break;
		case CONT_D_MAGIC:
			id+=2;
			contupdate[port]=contupdate_nothing;
			contread[port]=contread_magickey;
			contwrite[port]=contwrite_nothing;
			break;
		case CONT_D_ARK:
			id+=3;
			contupdate[port]=contupdate_arkanoid;
			contread[port]=contread_arkanoid;
			contwrite[port]=contwrite_arkanoid;
			break;
		
		default:
			id+=4;
			contupdate[port]=contupdate_nothing;
			contread[port]=contread_nothing;
			contwrite[port]=contwrite_nothing;
			break;
	}
	
	main_menu_radio(id,IDM_PORT_FIRST+port*IDM_PORT_SIZE,(IDM_PORT_FIRST+port*IDM_PORT_SIZE)+IDM_PORT_SIZE-1);
	
	input_affirm_ports();
}

void cont_user_insert(UINT id)
{
	int port=0,device=0;
	
	if (netplay_is_active()||movie_get_active_state()) return;
	
	switch (id) {
		case IDM_PORT1_JOY: port=0; device=CONT_D_J; break;
		case IDM_PORT1_HS: port=0; device=CONT_D_HS; break;
		case IDM_PORT1_MAGIC: port=0; device=CONT_D_MAGIC; break;
		case IDM_PORT1_ARK: port=0; device=CONT_D_ARK; break;
		case IDM_PORT1_NO: port=0; device=CONT_D_NOTHING; break;
		
		case IDM_PORT2_JOY: port=1; device=CONT_D_J; break;
		case IDM_PORT2_HS: port=1; device=CONT_D_HS; break;
		case IDM_PORT2_MAGIC: port=1; device=CONT_D_MAGIC; break;
		case IDM_PORT2_ARK: port=1; device=CONT_D_ARK; break;
		case IDM_PORT2_NO: port=1; device=CONT_D_NOTHING; break;
		
		default: return;
	}
	
	msx_wait();
	cont_insert(port,device,FALSE,TRUE);
	msx_wait_done();
}


/* state		size
port 1			1
port 2			1

curport			1
keyrow			12
keyextra		1
stateextra		1
sonypause_prev		1

gen_output		1/1
joy_state		1/1
hs_state		1/1
ark_shift_upd		2/2
ark_shift		2/2
ark_button		1/1
ark_bitcount		1/1

==			36
*/
#define STATE_VERSION	2
/* version history:
1: initial
2: improved gen_output (was gen_strobe) + added hypershot + added keyextra/stateextra/sonypause + added torikeshi/jikkou keys
*/
#define STATE_SIZE	36

int __fastcall cont_state_get_version(void)
{
	return STATE_VERSION;
}

int __fastcall cont_state_get_size(void)
{
	return STATE_SIZE;
}

void __fastcall cont_state_save(u8** s)
{
	STATE_SAVE_1(cont_port[CONT_P1]);	STATE_SAVE_1(cont_port[CONT_P2]);
	STATE_SAVE_1(curport);
	
	STATE_SAVE_C(keyrow,12);
	STATE_SAVE_1(keyextra);			STATE_SAVE_1(stateextra);
	STATE_SAVE_1(sonypause_prev);
	
	STATE_SAVE_1(gen_output[0]);		STATE_SAVE_1(gen_output[1]);
	STATE_SAVE_1(joy_state[0]);		STATE_SAVE_1(joy_state[1]);
	STATE_SAVE_1(hs_state[0]);		STATE_SAVE_1(hs_state[1]);
	STATE_SAVE_2(ark_shift_upd[0]);		STATE_SAVE_2(ark_shift_upd[1]);
	STATE_SAVE_2(ark_shift[0]);		STATE_SAVE_2(ark_shift[1]);
	STATE_SAVE_1(ark_button[0]);		STATE_SAVE_1(ark_button[1]);
	STATE_SAVE_1(ark_bitcount[0]);		STATE_SAVE_1(ark_bitcount[1]);
}

void __fastcall cont_state_load_cur(u8** s)
{
	STATE_LOAD_1(cont_port[CONT_P1]);	STATE_LOAD_1(cont_port[CONT_P2]);
	STATE_LOAD_1(curport);
	
	STATE_LOAD_C(keyrow,12);
	STATE_LOAD_1(keyextra);			STATE_LOAD_1(stateextra);
	STATE_LOAD_1(sonypause_prev);
	
	STATE_LOAD_1(gen_output[0]);		STATE_LOAD_1(gen_output[1]);
	STATE_LOAD_1(joy_state[0]);		STATE_LOAD_1(joy_state[1]);
	STATE_LOAD_1(hs_state[0]);		STATE_LOAD_1(hs_state[1]);
	STATE_LOAD_2(ark_shift_upd[0]);		STATE_LOAD_2(ark_shift_upd[1]);
	STATE_LOAD_2(ark_shift[0]);		STATE_LOAD_2(ark_shift[1]);
	STATE_LOAD_1(ark_button[0]);		STATE_LOAD_1(ark_button[1]);
	STATE_LOAD_1(ark_bitcount[0]);		STATE_LOAD_1(ark_bitcount[1]);
}

void __fastcall cont_state_load_1(u8** s)
{
	/* version 1 */
	STATE_LOAD_1(cont_port[CONT_P1]);	STATE_LOAD_1(cont_port[CONT_P2]);
	STATE_LOAD_1(curport);
	
	STATE_LOAD_C(keyrow,11);
	
	STATE_LOAD_1(gen_output[0]);		STATE_LOAD_1(gen_output[1]);
	STATE_LOAD_1(joy_state[0]);		STATE_LOAD_1(joy_state[1]);
	STATE_LOAD_2(ark_shift_upd[0]);		STATE_LOAD_2(ark_shift_upd[1]);
	STATE_LOAD_2(ark_shift[0]);		STATE_LOAD_2(ark_shift[1]);
	STATE_LOAD_1(ark_button[0]);		STATE_LOAD_1(ark_button[1]);
	STATE_LOAD_1(ark_bitcount[0]);		STATE_LOAD_1(ark_bitcount[1]);
	
	/* gen_output */
	gen_output[0]=(gen_output[0]&3)|(gen_output[0]>>2&4);
	gen_output[1]=(gen_output[1]>>2&3)|(gen_output[1]>>3&4);
	
	/* hypershot */
	hs_state[0]=hs_state[1]=0x3f;
	
	/* extra key/state */
	keyextra=stateextra=0;
	sonypause_prev=0xff;
	
	/* torikeshi/jikkou */
	keyrow[11]=0xff;
}

int __fastcall cont_state_load(int v,u8** s)
{
	switch (v) {
		case 1:
			cont_state_load_1(s);
			break;
		case STATE_VERSION:
			cont_state_load_cur(s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

/* movie	size

keyboard	12 (must)
keyextra	1  (must)
joystick	1+1
hypershot	1+1
arkanoid	2+2

MAX		17

*/
int __fastcall cont_movie_get_size(void)
{
	int port=2,size=13; /* keyboard+extra */
	
	/* ports */
	while (port--) {
		switch (cont_port[port]) {
			case CONT_D_J: case CONT_D_HS: size++; break;
			case CONT_D_ARK: size+=2; break;
			
			default: break;
		}
	}
	
	return size;
}

void __fastcall cont_movie_put(u8* c)
{
	int port=2,i=12; /* keyboard size */
	
	/* keyboard */
	memcpy(keyrow,c,12);
	keyextra=c[i++];
	
	/* ports */
	while (port--) {
		switch (cont_port[port]) {
			case CONT_D_J: joy_state[port]=c[i++]; break;
			case CONT_D_HS: hs_state[port]=c[i++]; break;
			case CONT_D_ARK: ark_button[port]=c[i]&2; ark_shift_upd[port]=(c[i]<<8|c[i+1])&0x1ff; i+=2; break;
			
			default: break;
		}
	}
}

void __fastcall cont_movie_get(u8* c)
{
	int port=2,i=12; /* keyboard size */
	
	/* keyboard */
	memcpy(c,keyrow,12);
	c[i++]=keyextra;
	
	/* ports */
	while (port--) {
		switch (cont_port[port]) {
			case CONT_D_J: c[i++]=joy_state[port]; break;
			case CONT_D_HS: c[i++]=hs_state[port]; break;
			case CONT_D_ARK: c[i++]=ark_button[port]|(ark_shift_upd[port]>>8&1); c[i++]=ark_shift_upd[port]&0xff; break;
			
			default: break;
		}
	}
}
