#include "global.h"
#include "io.h"
#include "mapper.h"
#include "vdp.h"
#include "cont.h"
#include "psg.h"
#include "sound.h"
#include "z80.h"
#include "crystal.h"
#include "state.h"


static u8 ppic=0; /* ppi port c */


/* keyclick sound, also used by some games (Beepertron, .. some Speccy ports, eg. Auf Wiedersehen Monty, Avenger, Jack the Nipper 2) */
#define CLICK_VOLUME 5567 /* max psg volume*~0.9278, measured on my MSX1 (i/o chip: Mitsubishi M5L8255AP-5) */

static int click_bc=0;
static int click_cc=0;
static int* click_dac;

static __inline__ void click_update(void)
{
	int i=(click_cc-z80_get_rel_cycles())/crystal->rel_cycle;
	const int sample=(ppic&0x80)?CLICK_VOLUME:0;
	
	if (i<=0) return;
	click_cc=z80_get_rel_cycles();
	
	while (i--) click_dac[click_bc++]=sample;
}

void io_set_click_buffer_counter(int bc)
{
	if (bc!=0&&bc<click_bc) {
		/* copy leftover to start */
		int i=bc,j=click_bc;
		while (i--) click_dac[i]=click_dac[--j];
	}
	
	click_bc=bc;
}

void io_init_click_sound(void) { click_dac=sound_create_dac(); }
int io_get_click_buffer_counter(void) { return click_bc; }
void io_new_frame(void) { click_cc+=crystal->frame; }
void io_end_frame(void) { click_update(); }


/* misc i/o port functions */
u8 __fastcall ioreadnull(void) { return 0xff; } /* $ff on my Philips VG-8235, open-bus (last opcode byte) OR 9 on my Canon V-20 */
void __fastcall iowritenull(u8 v) { ; }

u8 __fastcall io_read_key(void) { return cont_get_keyrow(ppic&0xf); }
u8 __fastcall io_read_ppic(void) { return ppic; }

void __fastcall io_write_ppic(u8 v)
{
	if ((v^ppic)&0x80) click_update(); /* click sound */
	
	ppic=v;
}

void __fastcall io_write_ppicontrol(u8 v)
{
	if (!(v&0x80)) {
		if (v&1) io_write_ppic(ppic|(1<<(v>>1&7)));
		else io_write_ppic(ppic&~(1<<(v>>1&7)));
	}
}


/* init/clean */
void io_poweron(void)
{
	click_cc=ppic=0;
}

void __fastcall io_setreadport(u8 p,fp_ioread f) { ioread[p]=f?f:ioreadnull; }
void __fastcall io_setwriteport(u8 p,fp_iowrite f) { iowrite[p]=f?f:iowritenull; }

void io_init(void)
{
	int i;
	for (i=0;i<0x100;i++) {
		io_setreadport(i,NULL);
		io_setwriteport(i,NULL);
	}
	
	/* set standard read ports */
	io_setreadport(0x98,vdp_read_data);
	io_setreadport(0x99,vdp_read_status);
	io_setreadport(0xa2,psg_read_data);
	io_setreadport(0xa8,mapper_read_pslot);
	io_setreadport(0xa9,io_read_key);
	io_setreadport(0xaa,io_read_ppic);
	
	/* set standard write ports */
	io_setwriteport(0x98,vdp_write_data);
	io_setwriteport(0x99,vdp_write_address);
	io_setwriteport(0xa0,psg_write_address);
	io_setwriteport(0xa1,psg_write_data);
	io_setwriteport(0xa8,mapper_write_pslot);
	io_setwriteport(0xaa,io_write_ppic);
	io_setwriteport(0xab,io_write_ppicontrol);
}

void io_clean(void)
{
	;
}


/* state		size
click_cc		4
ppic			1

==			5
*/
#define STATE_VERSION	2
/* version history:
1: initial
2: added tape (nothing changed here)
*/
#define STATE_SIZE	5

int __fastcall io_state_get_version(void)
{
	return STATE_VERSION;
}

int __fastcall io_state_get_size(void)
{
	return STATE_SIZE;
}

void __fastcall io_state_save(u8** s)
{
	STATE_SAVE_4(click_cc);
	STATE_SAVE_1(ppic);
}

void __fastcall io_state_load_cur(u8** s)
{
	STATE_LOAD_4(click_cc);
	STATE_LOAD_1(ppic);
}

int __fastcall io_state_load(int v,u8** s)
{
	switch (v) {
		case 1: case STATE_VERSION:
			io_state_load_cur(s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}
