/* Konami mappers (and SCC clones) */

/* --- Konami
	Konami megaROMs without SCC, eg.
	Gradius (Nemesis), Maze of Galious, Penguin Adventure, Shalom
	(seems like most Konami's SCC games work too, without SCC of course)
*/
static int konami_mask[2];

void __fastcall mapwrite_konami(u16 a,u8 v)
{
	/* last 3 pages switchable at $6000-$bfff */
	int slot=~pslot>>(a>>13&6)&1;
	
	cartbank[slot][a>>13]=v&konami_mask[slot];
	refresh_cb(slot);
}

static void mapinit_konami(int slot)
{
	int i;
	
	konami_mask[slot]=cartmask[slot];
	if (konami_mask[slot]<0x1f) konami_mask[slot]=0x1f; /* 256KB mapper, but allow hacks too if it turns out the ROM is larger */
	
	/* 4 pages of 8KB at $4000-$bfff */
	for (i=2;i<6;i++) cartbank[slot][i]=i-2;
	for (i=3;i<6;i++) mapwrite_std[slot+1][i]=mapwrite_konami;
}



/* --- Konami Game Master 2 (single game) */
static int gm2_is_sram[2][3];
static int gm2_sram_bank[2][3];

u8 __fastcall mapread_gm2_sram_s1(u16 a) { return cartsram[0][(a&0xfff)|gm2_sram_bank[0][(a>>13)-3]]; }
u8 __fastcall mapread_gm2_sram_s2(u16 a) { return cartsram[1][(a&0xfff)|gm2_sram_bank[1][(a>>13)-3]]; }
const fp_mapread mapread_gm2[2]={mapread_gm2_sram_s1,mapread_gm2_sram_s2};

void __fastcall mapwrite_gm2_std(u16,u8);
void __fastcall mapwrite_gm2_sram_s1(u16 a,u8 v) { if (a&0x1000) cartsram[0][(a&0xfff)|gm2_sram_bank[0][(a>>13)-3]]=v; else mapwrite_gm2_std(a,v); }
void __fastcall mapwrite_gm2_sram_s2(u16 a,u8 v) { if (a&0x1000) cartsram[1][(a&0xfff)|gm2_sram_bank[1][(a>>13)-3]]=v; else mapwrite_gm2_std(a,v); }
const fp_mapwrite mapwrite_gm2[2]={mapwrite_gm2_sram_s1,mapwrite_gm2_sram_s2};

static __inline__ void gm2_refresh(int slot)
{
	/* map sram */
	
	/* $6000-$9fff: read-only */
	if (gm2_is_sram[slot][0])	mapread_std[slot+1][3]=mapread_gm2[slot];
	else				mapread_std[slot+1][3]=lut_mapread[slot][3];
	if (gm2_is_sram[slot][1])	mapread_std[slot+1][4]=mapread_gm2[slot];
	else				mapread_std[slot+1][4]=lut_mapread[slot][4];
	
	/* $8000-$bfff: read/(part)write */
	if (gm2_is_sram[slot][2]) {	mapread_std[slot+1][5]=mapread_gm2[slot];	mapwrite_std[slot+1][5]=mapwrite_gm2[slot];	}
	else {				mapread_std[slot+1][5]=lut_mapread[slot][5];	mapwrite_std[slot+1][5]=mapwrite_gm2_std;	}
	
	mapper_refresh_pslot_read();
	mapper_refresh_pslot_write();
}

void __fastcall mapwrite_gm2_std(u16 a,u8 v)
{
	int slot=~pslot>>(a>>13&6)&1;
	int am=(a>>13)-3;
	
	if (a&0x1000) return;
	
	/* sram, 2*4K */
	gm2_is_sram[slot][am]=v&(cartmask[slot]+1);
	gm2_sram_bank[slot][am]=((v>>1&(cartmask[slot]+1))!=0)<<12;
	gm2_refresh(slot);
	
	/* similar to standard Konami */
	cartbank[slot][a>>13]=v&cartmask[slot];
	refresh_cb(slot);
}

static void mapclean_gm2(int slot)
{
	sram_save(slot,cartsram[slot],0x2000,NULL);
}

/* state			size
banks sram enabled		3*1
sram bank numbers		3*2

==				9
*/
#define STATE_VERSION_GM2	2 /* mapper.c */
/* version history:
1: initial
2: no changes (mapper.c)
*/
#define STATE_SIZE_GM2		9

static int __fastcall mapgetstatesize_gm2(int slot)
{
	return STATE_SIZE_GM2;
}

static void __fastcall mapsavestate_gm2(int slot,u8** s)
{
	int i=3;
	while (i--) {
		STATE_SAVE_1(gm2_is_sram[slot][i]);
		STATE_SAVE_2(gm2_sram_bank[slot][i]);
	}
}

static void __fastcall maploadstatecur_gm2(int slot,u8** s)
{
	int i=3;
	while (i--) {
		STATE_LOAD_1(gm2_is_sram[slot][i]);
		STATE_LOAD_2(gm2_sram_bank[slot][i]);
	}
	
	gm2_refresh(slot);
}

static int __fastcall maploadstate_gm2(int slot,int v,u8** s)
{
	switch (v) {
		case 1: case STATE_VERSION_GM2:
			maploadstatecur_gm2(slot,s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

static void mapinit_gm2(int slot)
{
	int i;
	
	if (mapclean[slot]==mapclean_nothing) {
		sram_load(slot,&cartsram[slot],0x2000,NULL,0xff);
		mapclean[slot]=mapclean_gm2;
	}
	
	mapgetstatesize[slot]=mapgetstatesize_gm2;
	mapsavestate[slot]=mapsavestate_gm2;
	maploadstatecur[slot]=maploadstatecur_gm2;
	maploadstate[slot]=maploadstate_gm2;
	
	for (i=0;i<3;i++) gm2_is_sram[slot][i]=FALSE;
	gm2_refresh(slot);
	
	/* similar to standard Konami */
	for (i=2;i<6;i++) cartbank[slot][i]=i-2;
	for (i=3;i<6;i++) mapwrite_std[slot+1][i]=mapwrite_gm2_std;
}



/* --- Konami SCC
	Konami megaROMs with SCC, eg.
	F1 Spirit, Gofer no Yabou 2 (Nemesis 3), Gradius 2, King's Valley 2, Parodius, Salamander
*/
static int scc_enabled[2];
static _scc* scc_scc[2]={NULL,NULL};

u8 __fastcall mapread_konamiscc(u16 a)
{
	int slot=~pslot>>4&1;
	
	/* read from SCC */
	if (scc_enabled[slot]&&a&0x1800) return scc_read(scc_scc[slot],a);
	
	/* default */
	return cartmap[slot][4][a&0x1fff];
}

void __fastcall mapwrite_konamiscc(u16 a,u8 v)
{
	/* all pages switchable at $4000-$bfff */
	int slot=~pslot>>(a>>13&6)&1;
	
	if ((a&0xf000)==0x9000) {
		if (a&0x800) {
			/* write to SCC */
			if (scc_enabled[slot]) scc_write(scc_scc[slot],a,v);
			return;
		}
		
		scc_enabled[slot]=(v&0x3f)==0x3f;
	}
	
	if ((a&0x1800)==0x1000) {
		/* bankswitch */
		cartbank[slot][a>>13]=v&cartmask[slot];
		refresh_cb(slot);
	}
}

static void mapnf_konamiscc(int slot) { scc_new_frame(scc_scc[slot]); }
static void mapef_konamiscc(int slot) { scc_end_frame(scc_scc[slot]); }
static void mapsetbc_konamiscc(int slot,int bc) { scc_set_buffer_counter(scc_scc[slot],bc); }

static void mapclean_konamiscc(int slot)
{
	scc_clean(scc_scc[slot]);
	scc_scc[slot]=NULL;
}

/* state: size 1 (enabled) + scc */
#define STATE_VERSION_KSCC	2 /* mapper.c */
/* version history:
1: initial
2: no changes (mapper.c)
*/
#define STATE_SIZE_KSCC		1 /* + SCC */

static int __fastcall mapgetstatesize_konamiscc(int slot)
{
	return STATE_SIZE_KSCC+scc_state_get_size();
}

static void __fastcall mapsavestate_konamiscc(int slot,u8** s)
{
	STATE_SAVE_1(scc_enabled[slot]);
	
	scc_state_save(scc_scc[slot],s);
}

static void __fastcall maploadstatecur_konamiscc_inc(int slot,u8** s)
{
	STATE_LOAD_1(scc_enabled[slot]);
}

static void __fastcall maploadstatecur_konamiscc(int slot,u8** s)
{
	maploadstatecur_konamiscc_inc(slot,s);
	
	scc_state_load_cur(scc_scc[slot],s);
}

static int __fastcall maploadstate_konamiscc(int slot,int v,u8** s)
{
	switch (v) {
		case 1:
			maploadstatecur_konamiscc_inc(slot,s);
			scc_state_load(scc_scc[slot],v,s);
			break;
		case STATE_VERSION_KSCC:
			maploadstatecur_konamiscc(slot,s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

static void mapinit_konamiscc(int slot)
{
	int i;
	
	/* 4 pages of 8KB at $4000-$bfff */
	for (i=2;i<6;i++) {
		cartbank[slot][i]=i-2;
		mapwrite_std[slot+1][i]=mapwrite_konamiscc;
	}
	mapread_std[slot+1][4]=mapread_konamiscc;
	
	if (mapclean[slot]==mapclean_nothing) {
		scc_scc[slot]=scc_init(SCC_VF_DEFAULT,SCC_MODEL);
		mapclean[slot]=mapclean_konamiscc;
	}
	
	scc_poweron(scc_scc[slot]);
	scc_enabled[slot]=FALSE;
	
	mapnf[slot]=mapnf_konamiscc;
	mapef[slot]=mapef_konamiscc;
	mapsetbc[slot]=mapsetbc_konamiscc;
	
	mapgetstatesize[slot]=mapgetstatesize_konamiscc;
	mapsavestate[slot]=mapsavestate_konamiscc;
	maploadstatecur[slot]=maploadstatecur_konamiscc;
	maploadstate[slot]=maploadstate_konamiscc;
}



/* --- Konami Sound Cartridge
	SCC-I with 128KB RAM, supported by the Game Collections (see below) (and some hacked ROMs)
	1: 1:Knightmare(YES), 2:Antarctic Adventure(YES), 3:Yie ar Kung-Fu(YES), 4:Yie ar Kung-Fu 2(YES), 5:King's Valley(NO)
	2: 1:Konami's Boxing(YES), 2:Konami's Tennis(NO), 3:Video Hustler(YES), 4:Hyper Olympic 1(NO), 5:Hyper Sports 2(YES)
	3: 1:Twinbee(YES), 2:Super Cobra(YES), 3:Sky Jaguar(YES), 4:Time Pilot(NO), 5:Nemesis(YES)
	4: 1:Konami's Soccer(YES), 2:Konami's Ping Pong(YES), 3:Konami's Golf(NO), 4:Hyper Olympic 2(NO), 5:Hyper Sports 3(YES)
	S: 1.1:Pippols(YES), 1.2:Hyper Rally(YES), 1.3:Road Fighter(YES) (other choices are MSX2-only)
*/
static int scci_mode[2];
static int scci_bank[2][4];
static int scci_we[2][4]; /* write enabled */
static u8 scci_ram[2][0x20000]; /* 128KB */

u8 __fastcall mapread_konamiscci(u16 a)
{
	int slot=~pslot>>(a>>13&6)&1;
	
	/* read from SCC-I, SCC mode */
	if ((a&0xf800)==0x9800&&scc_enabled[slot]&1&&!(scci_mode[slot]&0x20)) return scc_read(scc_scc[slot],a);
	
	/* read from SCC-I, SCC-I mode */
	if ((a&0xf800)==0xb800&&scc_enabled[slot]&0x80&scci_mode[slot]<<2) return scci_read(scc_scc[slot],a);
	
	/* read from ram */
	return scci_ram[slot][scci_bank[slot][(a>>13)-2]|(a&0x1fff)];
}

void __fastcall mapwrite_konamiscci(u16 a,u8 v)
{
	int page=a>>13;
	int slot=~pslot>>(page&6)&1;
	page-=2;
	
	if ((a&0xfffe)==0xbffe) {
		/* mode register */
		scci_mode[slot]=v;
		
		scci_we[slot][0]=v&0x11;
		scci_we[slot][1]=v&0x12;
		scci_we[slot][2]=(v>>3&v&4)|(v&0x10);
		scci_we[slot][3]=v&0x10;
		
		return;
	}
	
	if (scci_we[slot][page]) {
		/* write to RAM */
		scci_ram[slot][scci_bank[slot][page]|(a&0x1fff)]=v;
	}
	else {
		switch (a&0xf800) {
			case 0x9000:
				/* enable SCC mode */
				scc_enabled[slot]=(scc_enabled[slot]&0x80)|((v&0x3f)==0x3f);
				break;
			case 0x9800:
				/* write to SCC-I, SCC mode */
				if (scc_enabled[slot]&1&&!(scci_mode[slot]&0x20)) scc_write(scc_scc[slot],a,v);
				break;
			case 0xb000:
				/* enable SCC-I mode */
				scc_enabled[slot]=(scc_enabled[slot]&1)|(v&0x80);
				break;
			case 0xb800:
				/* write to SCC-I, SCC-I mode */
				if (scc_enabled[slot]&0x80&scci_mode[slot]<<2) scci_write(scc_scc[slot],a,v);
				break;
			
			default: break;
		}
		
		if ((a&0x1800)==0x1000) {
			/* bankswitch */
			scci_bank[slot][page]=v<<13&0x1ffff;
		}
	}
}

/* state		size
enabled			1
mode register		1
banks			4*3
banks write enable	4
ram			0x20000 (128KB)

==			0x20012 (+SCC)
*/
#define STATE_VERSION_KSCCI	2 /* mapper.c */
/* version history:
1: (didn't exist yet)
2: initial
*/
#define STATE_SIZE_KSCCI	0x20012 /* + SCC */

static int __fastcall mapgetstatesize_konamiscci(int slot)
{
	return STATE_SIZE_KSCCI+scc_state_get_size();
}

static void __fastcall mapsavestate_konamiscci(int slot,u8** s)
{
	int i=4;
	while (i--) {
		STATE_SAVE_3(scci_bank[slot][i]);
		STATE_SAVE_1(scci_we[slot][i]);
	}
	
	STATE_SAVE_1(scc_enabled[slot]);
	STATE_SAVE_1(scci_mode[slot]);
	
	scc_state_save(scc_scc[slot],s);
}

static void __fastcall maploadstatecur_konamiscci(int slot,u8** s)
{
	int i=4;
	while (i--) {
		STATE_LOAD_3(scci_bank[slot][i]);
		STATE_LOAD_1(scci_we[slot][i]);
	}
	
	STATE_LOAD_1(scc_enabled[slot]);
	STATE_LOAD_1(scci_mode[slot]);
	
	scc_state_load_cur(scc_scc[slot],s);
}

static int __fastcall maploadstate_konamiscci(int slot,int v,u8** s)
{
	switch (v) {
		case STATE_VERSION_KSCC:
			maploadstatecur_konamiscci(slot,s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

static void mapinit_konamiscci(int slot)
{
	int i;
	
	/* 4 pages of 8KB at $4000-$bfff */
	for (i=2;i<6;i++) {
		mapread_std[slot+1][i]=mapread_konamiscci;
		mapwrite_std[slot+1][i]=mapwrite_konamiscci;
	}
	
	for (i=0;i<4;i++) {
		scci_we[slot][i]=FALSE;
		scci_bank[slot][i]=i<<13;
	}
	
	if (mapclean[slot]==mapclean_nothing) {
		scc_scc[slot]=scc_init(SCC_VF_DEFAULT,SCCI_MODEL);
		mapclean[slot]=mapclean_konamiscc;
		
		/* file -> RAM (normally unsupported) */
		for (i=0;i<0x10;i++) memcpy(scci_ram[slot]+0x2000*i,cart[slot][i],0x2000);
	}
	
	scc_poweron(scc_scc[slot]);
	scc_enabled[slot]=0;
	scci_mode[slot]=cartpages[slot]?0x20:0; /* pre-init if a file was loaded */
	
	mapnf[slot]=mapnf_konamiscc;
	mapef[slot]=mapef_konamiscc;
	mapsetbc[slot]=mapsetbc_konamiscc;
	
	mapgetstatesize[slot]=mapgetstatesize_konamiscci;
	mapsavestate[slot]=mapsavestate_konamiscci;
	maploadstatecur[slot]=maploadstatecur_konamiscci;
	maploadstate[slot]=maploadstate_konamiscci;
}



/* --- Manuel Mega Flash ROM SCC
	Lotus F3, ..
*/
static int mfrscc_sector[2][4];
static int mfrscc_bank[2][4];
static _am29f* mfrscc_chip[2]={NULL,NULL};

u8 __fastcall mapread_mfrscc(u16 a)
{
	/* like Konami SCC */
	int bank=a>>13;
	int slot=~pslot>>(bank&6)&1;
	bank-=2;
	
	if (scc_enabled[slot]&&(a&0xf800)==0x9800) return scc_read(scc_scc[slot],a);
	
	/* from flash */
	mfrscc_chip[slot]->readsector=mfrscc_chip[slot]->writesector=mfrscc_sector[slot][bank];
	return am29f_read(mfrscc_chip[slot],mfrscc_bank[slot][bank]|(a&0x1fff));
}

void __fastcall mapwrite_mfrscc(u16 a,u8 v)
{
	/* like Konami SCC */
	int bank=a>>13;
	int slot=~pslot>>(bank&6)&1;
	bank-=2;
	
	if ((a&0xf000)==0x9000) {
		if (a&0x800) {
			/* write to SCC */
			if (scc_enabled[slot]) scc_write(scc_scc[slot],a,v);
		}
		else scc_enabled[slot]=(v&0x3f)==0x3f;
	}
	
	if ((a&0x1800)==0x1000) {
		/* bankswitch */
		int s=v<<13;
		mfrscc_sector[slot][bank]=s&0x70000;
		mfrscc_bank[slot][bank]=s&0xf000;
	}
	
	/* to flash */
	mfrscc_chip[slot]->readsector=mfrscc_chip[slot]->writesector=mfrscc_sector[slot][bank];
	am29f_write(mfrscc_chip[slot],mfrscc_bank[slot][bank]|(a&0x1fff),v);
}


static void mapclean_mfrscc(int slot)
{
	mapclean_konamiscc(slot);
	sram_save(slot,mfrscc_chip[slot]->data,0x80000,NULL);
	am29f_clean(mfrscc_chip[slot]);
	mfrscc_chip[slot]=NULL;
}

/* state		size
sector			3*4
bank			2*4
SCC enabled		1

(+SCC+amf29f)

==			21
*/
#define STATE_VERSION_MFRSCC	2 /* mapper.c */
/* version history:
1: initial
2: no changes (mapper.c)
*/
#define STATE_SIZE_MFRSCC	21

static int __fastcall mapgetstatesize_mfrscc(int slot)
{
	return STATE_SIZE_MFRSCC+scc_state_get_size()+am29f_state_get_size();
}

static void __fastcall mapsavestate_mfrscc(int slot,u8** s)
{
	int i=4;
	while (i--) {
		STATE_SAVE_3(mfrscc_sector[slot][i]);
		STATE_SAVE_2(mfrscc_bank[slot][i]);
	}
	STATE_SAVE_1(scc_enabled[slot]);
	
	scc_state_save(scc_scc[slot],s);
	am29f_state_save(mfrscc_chip[slot],s);
}

static void __fastcall maploadstatecur_mfrscc_inc(int slot,u8** s)
{
	int i=4;
	while (i--) {
		STATE_LOAD_3(mfrscc_sector[slot][i]);
		STATE_LOAD_2(mfrscc_bank[slot][i]);
	}
	STATE_LOAD_1(scc_enabled[slot]);
}

static void __fastcall maploadstatecur_mfrscc(int slot,u8** s)
{
	maploadstatecur_mfrscc_inc(slot,s);
	
	scc_state_load_cur(scc_scc[slot],s);
	am29f_state_load_cur(mfrscc_chip[slot],s);
}

static int __fastcall maploadstate_mfrscc(int slot,int v,u8** s)
{
	switch (v) {
		case 1:
			maploadstatecur_mfrscc_inc(slot,s);
			scc_state_load(scc_scc[slot],v,s);
			am29f_state_load(mfrscc_chip[slot],v,s);
			break;
		case STATE_VERSION_MFRSCC:
			maploadstatecur_mfrscc(slot,s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

static void mapinit_mfrscc(int slot)
{
	int i;
	
	if (mapclean[slot]==mapclean_nothing) {
		mfrscc_chip[slot]=am29f_init(FALSE);
		for (i=0;i<8;i++) mfrscc_chip[slot]->unprotect[i]=TRUE;
		
		if (!sram_load(slot,&mfrscc_chip[slot]->data,0x80000,NULL,0xff)) {
			/* upload game to flash */
			for (i=0;i<0x40;i++) memcpy(mfrscc_chip[slot]->data+0x2000*i,cart[slot][i],0x2000);
		}
		
		scc_scc[slot]=scc_init(SCC_VF_MFR,SCC_MODEL);
		mapclean[slot]=mapclean_mfrscc;
	}
	mfrscc_chip[slot]->readsector=mfrscc_chip[slot]->writesector=0;
	am29f_reset(mfrscc_chip[slot]);
	scc_poweron(scc_scc[slot]);
	scc_enabled[slot]=FALSE;
	
	for (i=0;i<4;i++) {
		mfrscc_sector[slot][i]=0;
		mfrscc_bank[slot][i]=i<<13;
	}
	
	for (i=2;i<6;i++) {
		mapwrite_std[slot+1][i]=mapwrite_mfrscc;
		mapread_std[slot+1][i]=mapread_mfrscc;
	}
	
	mapnf[slot]=mapnf_konamiscc;
	mapef[slot]=mapef_konamiscc;
	mapsetbc[slot]=mapsetbc_konamiscc;
	
	mapgetstatesize[slot]=mapgetstatesize_mfrscc;
	mapsavestate[slot]=mapsavestate_mfrscc;
	maploadstatecur[slot]=maploadstatecur_mfrscc;
	maploadstate[slot]=maploadstate_mfrscc;
}



/* --- Vincent DSK2ROM (uses Konami SCC)
	any disk file, read-only
*/
#define DSK2ROM_FILE "dsk2rom.rom"
static u32 dsk2rom_crcprev[2]={0,0};

static void mapclean_dsk2rom(int slot)
{
	int i;
	
	/* restructure (backwards) */
	MEM_CLEAN(cart[slot][0]); MEM_CLEAN(cart[slot][1]);
	for (i=2;i<cartpages[slot];i++) cart[slot][i-2]=cart[slot][i];
	cart[slot][cartpages[slot]-1]=cart[slot][cartpages[slot]-2]=dummy;
	cartpages[slot]-=2;
	for (cartmask[slot]=1;cartmask[slot]<cartpages[slot];cartmask[slot]<<=1) { ; }
	cartmask[slot]--;
	
	cartcrc[slot]=dsk2rom_crcprev[slot];
	
	/* konamiscc part */
	mapclean_konamiscc(slot);
}

static void dsk2rom_open(int slot)
{
	u8* dsk2rom[2];
	MEM_CREATE(dsk2rom[0],0x2000); MEM_CREATE(dsk2rom[1],0x2000);
	
	file_setfile(&file->appdir,DSK2ROM_FILE,NULL,NULL);
	if (file_open()&&file->size==0x4000&&file_read(dsk2rom[0],0x2000)&&file_read(dsk2rom[1],0x2000)) {
		file_close();
		if (cartpages[slot]!=0&&(cartpages[slot]+2)<=ROM_MAXPAGES&&memcmp(cart[slot][0],dsk2rom[0],0x2000)!=0&&memcmp(cart[slot][1],dsk2rom[1],0x2000)!=0) {
			/* restructure */
			int i=cartpages[slot];
			while (i--) cart[slot][i+2]=cart[slot][i];
			cart[slot][0]=dsk2rom[0]; cart[slot][1]=dsk2rom[1];
			dsk2rom[0]=dsk2rom[1]=NULL;
			cartpages[slot]+=2;
			for (cartmask[slot]=1;cartmask[slot]<cartpages[slot];cartmask[slot]<<=1) { ; }
			cartmask[slot]--;
			
			dsk2rom_crcprev[slot]=cartcrc[slot];
			cartcrc[slot]=(cartcrc[slot]<<1|(cartcrc[slot]>>31&1))^file->crc32; /* combine crc with dsk2rom crc */
			
			mapclean[slot]=mapclean_dsk2rom;
			
			/* konamiscc part */
			scc_scc[slot]=scc_init(SCC_VF_DEFAULT,SCC_MODEL);
		}
	}
	else {
		file_close();
		LOG(LOG_MISC|LOG_WARNING|LOG_TYPE(LT_SLOT1FAIL+slot),"slot %d: couldn't open %s",slot+1,DSK2ROM_FILE);
	}
	
	MEM_CLEAN(dsk2rom[0]); MEM_CLEAN(dsk2rom[1]);
}

static void mapinit_dsk2rom(int slot)
{
	if (mapclean[slot]==mapclean_nothing) dsk2rom_open(slot); /* onetime init */
	
	mapinit_konamiscc(slot);
}



/* --- Konami Synthesizer (single game) */
#define KSYN_VOL_FACTOR 12 /* guessed */
static int* konamisyn_dac[2];
static int konamisyn_output[2];
static int konamisyn_cc[2];
static int konamisyn_bc[2];

static __inline__ void konamisyn_update(int slot)
{
	int i=(konamisyn_cc[slot]-z80_get_rel_cycles())/crystal->rel_cycle;
	if (i<=0) return;
	konamisyn_cc[slot]=z80_get_rel_cycles();
	
	while (i--) konamisyn_dac[slot][konamisyn_bc[slot]++]=konamisyn_output[slot];
}

void __fastcall mapwrite_konamisyn(u16 a,u8 v)
{
	/* 8 bit PCM at $4000-$7fff */
	if (~a&0x10) {
		int slot=~pslot>>2&1;
		konamisyn_update(slot);
		konamisyn_output[slot]=v*KSYN_VOL_FACTOR;
	}
}

static void mapnf_konamisyn(int slot) { konamisyn_cc[slot]+=crystal->frame; }
static void mapef_konamisyn(int slot) { konamisyn_update(slot); }

static void mapsetbc_konamisyn(int slot,int bc)
{
	if (bc!=0&&bc<konamisyn_bc[slot]) {
		/* copy leftover to start */
		int i=bc,j=konamisyn_bc[slot];
		while (i--) konamisyn_dac[slot][i]=konamisyn_dac[slot][--j];
	}
	
	konamisyn_bc[slot]=bc;
}

static void mapclean_konamisyn(int slot)
{
	sound_clean_dac(konamisyn_dac[slot]);
	konamisyn_dac[slot]=NULL;
}

/* state			size
cc				4
output				4

==				8
*/
#define STATE_VERSION_KSYN	2 /* mapper.c */
/* version history:
1: initial
2: no changes (mapper.c)
*/
#define STATE_SIZE_KSYN		8

static int __fastcall mapgetstatesize_konamisyn(int slot)
{
	return STATE_SIZE_KSYN;
}

static void __fastcall mapsavestate_konamisyn(int slot,u8** s)
{
	STATE_SAVE_4(konamisyn_cc[slot]);
	STATE_SAVE_4(konamisyn_output[slot]);
}

static void __fastcall maploadstatecur_konamisyn(int slot,u8** s)
{
	STATE_LOAD_4(konamisyn_cc[slot]);
	STATE_LOAD_4(konamisyn_output[slot]);
}

static int __fastcall maploadstate_konamisyn(int slot,int v,u8** s)
{
	switch (v) {
		case 1: case STATE_VERSION_KSYN:
			maploadstatecur_konamisyn(slot,s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}

static void mapinit_konamisyn(int slot)
{
	if (mapclean[slot]==mapclean_nothing) {
		konamisyn_dac[slot]=sound_create_dac();
		
		mapclean[slot]=mapclean_konamisyn;
		mapnf[slot]=mapnf_konamisyn;
		mapef[slot]=mapef_konamisyn;
		mapsetbc[slot]=mapsetbc_konamisyn;
		mapgetstatesize[slot]=mapgetstatesize_konamisyn;
		mapsavestate[slot]=mapsavestate_konamisyn;
		maploadstatecur[slot]=maploadstatecur_konamisyn;
		maploadstate[slot]=maploadstate_konamisyn;
	}
	
	konamisyn_output[slot]=0;
	
	/* sync */
	konamisyn_cc[slot]=psg_get_cc();
	konamisyn_bc[slot]=psg_get_buffer_counter();
	
	mapwrite_std[slot+1][2]=mapwrite_konamisyn;
	mapwrite_std[slot+1][3]=mapwrite_konamisyn;
	
	mapinit_4000(slot); /* mapper_simple.c */
}
