/*

Lookup table for GoodMSX mapper types, good dumps, and some free games. I don't like this much (eg. hack a ROM and it
won't be detected properly anymore), but as MSX ROMs don't include metadata, this is a must (or at least convenient).
No need for (most) small ROMs, ASCII8, DSK2ROM, or R-Type; they're autodetected by size/start address.

*/

typedef struct {
	u32 crc;
	int type;
} _lutmaptype;

static const _lutmaptype lutmaptype[]={
	/* ASCII 8 + SRAM */
	{ 0x27FD8F9A,	CARTTYPE_ASCII8S },	/* Deep Dungeon 1 (1988) (Scap Trust) (J) */
	{ 0x101DB19C,	CARTTYPE_ASCII8S },	/* Deep Dungeon 2 (1988) (Scap Trust) (J) */
	{ 0xB8FC19A4,	CARTTYPE_ASCII8S },	/* Psychic War (1987) (Kogado) (J) */
	{ 0x0E691C47,	CARTTYPE_ASCII8S },	/* Syougun (1987) (Jaleco) (J) */
	{ 0xD640DEAF,	CARTTYPE_ASCII8S },	/* Xanadu (1987) (Falcom) (J) */
	
	{ 0x5ACBD591,	CARTTYPE_ASCII8S },	/* Dires (1987) (Telenet) (J) */
	{ 0xA8E9AC9B,	CARTTYPE_ASCII8S },	/* Erusurid - Elslid (1988) (NCS) (J) */
	{ 0x6AC18445,	CARTTYPE_ASCII8S },	/* Heroes Of The Lance (1991) (Pony Cannon) (J) */
	{ 0x459478CB,	CARTTYPE_ASCII8S },	/* Japanese MSX-Write II (1988) (Ascii) (J) */
	{ 0x167A53DA,	CARTTYPE_ASCII8S },	/* Kisei (1987) (Sony) (J) */
	{ 0xF640670C,	CARTTYPE_ASCII8S },	/* Shougi Sinan 2 (1988) (Pony Cannon) (J) */
	{ 0x39E1C37A,	CARTTYPE_ASCII8S },	/* Siryousensen - War Of The Dead (1987) (Victor) (J) */
	{ 0x37B9E10F,	CARTTYPE_ASCII8S },	/* Ultima Exodus (1987) (Pony Cannon) (J) */
	{ 0x08A5CA60,	CARTTYPE_ASCII8S },	/* Wizardry (1988) (Ascii) (J) */
	
	/* Koei */
	{ 0x4EAD5098,	CARTTYPE_KOEI },	/* Genghis Khan (1988) (Koei) (J) */
	{ 0x3AA33A30,	CARTTYPE_KOEI },	/* Nobunaga no Yabou - Zenkokuhan (1987) (Koei) (J) */
	{ 0x10621440,	CARTTYPE_KOEI },	/* Sangokusi 1 (1986) (Koei) (J) */
	{ 0x803CC690,	CARTTYPE_KOEI },	/* Sangokusi 1 (1986) (Koei) (J) [a1] */
	
	{ 0xEC7A8012,	CARTTYPE_KOEI },	/* Daikoukai Jidai (1990) (Koei) (J) */
	{ 0x413D4700,	CARTTYPE_KOEI },	/* Europe War (1992) (Koei) (J) */
	{ 0x075AB646,	CARTTYPE_KOEI },	/* Genchohisi (1992) (Koei) (J) */
	{ 0x3B5B3FD8,	CARTTYPE_KOEI },	/* Genghis Khan (1988) (Koei) (J) */
	{ 0x881B3849,	CARTTYPE_KOEI },	/* Inindo Tado Nobunaga (1991) (Koei) (J) */
	{ 0x8BC34DCA,	CARTTYPE_KOEI },	/* Isin no Arashi (1989) (Koei) (J) */
	{ 0xC7521809,	CARTTYPE_KOEI },	/* Lempereur (1990) (Koei) (J) */
	{ 0x0FF8E2C7,	CARTTYPE_KOEI },	/* Nobunaga no Yabou - Bushouhuunroku (1991) (Koei) (J) */
	{ 0x69C11D81,	CARTTYPE_KOEI },	/* Nobunaga no Yabou - Senkokugunyuden (1989) (Koei) (J) */
	{ 0x653F1C73,	CARTTYPE_KOEI },	/* Nobunaga no Yabou - Zenkokuhan (1987) (Koei) (J) */
	{ 0xFB9E42E1,	CARTTYPE_KOEI },	/* Royal Blood (1991) (Koei) (J) */
	{ 0xB83AA6F5,	CARTTYPE_KOEI },	/* Sangokusi 1 (1988) (Koei) (J) */
	{ 0xC89B93B5,	CARTTYPE_KOEI },	/* Sangokusi 2 (1991) (Koei) (J) */
	{ 0x187D3A21,	CARTTYPE_KOEI },	/* Suikoden (1989) (Koei) (J) */
	{ 0xC31F459C,	CARTTYPE_KOEI },	/* Teitoku no Ketsudan (1991) (Koei) (J) */
	
	
	/* ASCII 16 */
	{ 0xC6FC7BD7,	CARTTYPE_ASCII16 },	/* Alien 2 (1987) (Square) (J) */
	{ 0x67BF8337,	CARTTYPE_ASCII16 },	/* Black Onyx 2, The (1986) (Ascii) (J) */
	{ 0xE040E8A1,	CARTTYPE_ASCII16 },	/* Borfesu (1987) (Xtal Soft) (J) */
	{ 0xD83966D0,	CARTTYPE_ASCII16 },	/* Craze (1988) (Heart Soft) (J) */
	{ 0x2526E568,	CARTTYPE_ASCII16 },	/* Dungeon Hunter (1989) (Ascii) (J) */
	{ 0x47EC57DA,	CARTTYPE_ASCII16 },	/* Dynamite Bowl (1988) (Toshiba Emi) (J) */
	{ 0x99EF860E,	CARTTYPE_ASCII16 },	/* Gall Force (1986) (Sony) (J) */
	{ 0x5EAC55DF,	CARTTYPE_ASCII16 },	/* Golvellius (1987) (Compile) (J) */
	{ 0xD6465702,	CARTTYPE_ASCII16 },	/* Jagur (1987) (Hudson) (J) */
	{ 0xA0481321,	CARTTYPE_ASCII16 },	/* King's Knight (1986) (Square) (J) */
	{ 0xA3B2FE71,	CARTTYPE_ASCII16 },	/* Knither Special (1987) (Dempa) (J) */
	{ 0xD446BA1E,	CARTTYPE_ASCII16 },	/* Magunam - Kiki Ippatsu (1988) (Toshiba Emi) (J) */
	{ 0x2C2020A0,	CARTTYPE_ASCII16 },	/* Meikyuu no Tobira - Gate of Labyrinth (1987) (Dempa) (J) */
	{ 0x2F8065EB,	CARTTYPE_ASCII16 },	/* Meikyuu Shinwa (1986) (Hal) (J) (Eggerland 2) */
	{ 0x788637D5,	CARTTYPE_ASCII16 },	/* Robowres 2001 (1987) (Micronet) (J) */
	{ 0x387C1DE7,	CARTTYPE_ASCII16 },	/* Romancia (1986) (Falcom) (J) */
	{ 0xF00B4BBC,	CARTTYPE_ASCII16 },	/* Super Pierrot (1987) (Taito) (J) */
	{ 0xB9B69A43,	CARTTYPE_ASCII16 },	/* Tumego 120 (1987) (Champion Soft) (J) */
	{ 0xFEA70207,	CARTTYPE_ASCII16 },	/* Vaxol (1987) (Heart Soft) (J) */
	
	{ 0xF5174422,	CARTTYPE_ASCII16 },	/* A-Class Mahjong (1988) (Pony Cannon) (J) */
	{ 0xD2EE4051,	CARTTYPE_ASCII16 },	/* Acrojet (1988) (System Soft) (J) */
	{ 0xC29EDD84,	CARTTYPE_ASCII16 },	/* Aleste (1988) (Compile) (J) */
	{ 0xC6F74489,	CARTTYPE_ASCII16 },	/* Andorogynus (1988) (Telenet) (J) */
	{ 0xAE3167EE,	CARTTYPE_ASCII16 },	/* Arctic (1989) (Pony Cannon) (J) */
	{ 0x9935E9BD,	CARTTYPE_ASCII16 },	/* Dynamite Bowl (1987) (Toshiba Emi) (J) */
	{ 0x356B7D2A,	CARTTYPE_ASCII16 },	/* Family Stadium (1989) (Namcot) (J) */
	{ 0x1BDFAEEB,	CARTTYPE_ASCII16 },	/* Fantasy Zone 2 (1987) (Pony Cannon) (J) */
	{ 0xFBDAFE4E,	CARTTYPE_ASCII16 },	/* Gakuen Monogatari - High School Story (1988) (Great) (J) */
	{ 0x33C37313,	CARTTYPE_ASCII16 },	/* Garakuta (1988) (Wachi Electronics) (J) */
	{ 0x28094960,	CARTTYPE_ASCII16 },	/* Girly Block (1988) (Telenet) (J) */
	{ 0x8DF56D42,	CARTTYPE_ASCII16 },	/* Hacker (1988) (Pony Cannon) (J) */
	{ 0xF35DF283,	CARTTYPE_ASCII16 },	/* Hole In One Special (1987) (Hal) (J) */
	{ 0x4708BDD1,	CARTTYPE_ASCII16 },	/* Hole In One Special (1987) (Hal) (J) [a1] */
	{ 0xA9A66075,	CARTTYPE_ASCII16 },	/* Ikari (1987) (SNK) (J) */
	{ 0x053AA66E,	CARTTYPE_ASCII16 },	/* Kenpelen Chess (1988) (Pony Cannon) (J) */
	{ 0xFB348DE6,	CARTTYPE_ASCII16 },	/* Koronis Rift (1987) (Pony Cannon) (J) */
	{ 0x17BAF3ED,	CARTTYPE_ASCII16 },	/* Makaijima - Higemaru (1987) (Capcom) (J) */
	{ 0xF7A9591D,	CARTTYPE_ASCII16 },	/* Malaya no Hibou (1988) (Pony Cannon) (J) */
	{ 0xF2544E2E,	CARTTYPE_ASCII16 },	/* Ogre (1986) (System Soft) (J) */
	{ 0xB1B24DA1,	CARTTYPE_ASCII16 },	/* Penguin-Kun Wars 2 (1988) (Ascii) (J) */
	{ 0x17F9EE7F,	CARTTYPE_ASCII16 },	/* Romancia (1986) (Falcom) (J) */
	{ 0x3B815DC6,	CARTTYPE_ASCII16 },	/* Shanghai (1987) (System Soft) (J) */
	{ 0x4A3ABB65,	CARTTYPE_ASCII16 },	/* Strategic Mars (1989) (dB-Soft) (J) */
	{ 0xB2392EF3,	CARTTYPE_ASCII16 },	/* Super Rambo Special (1986) (Pack In Video) (J) */
	{ 0x5F354E65,	CARTTYPE_ASCII16 },	/* The Cockpit (1987) (Taito) (J) */
	{ 0x8EE8212C,	CARTTYPE_ASCII16 },	/* The Return Of Ishtar (1988) (Namcot) (J) */
	{ 0x94EBC319,	CARTTYPE_ASCII16 },	/* Victorious Nine 2 (1987) (Taito) (J) */
	{ 0xAEA8D064,	CARTTYPE_ASCII16 },	/* Woody Poco (1987) (dB-Soft) (J) */
	{ 0x97591725,	CARTTYPE_ASCII16 },	/* Xevious - Fardraut Saga (1989) (Namcot) (J) */
	{ 0xD153B2F2,	CARTTYPE_ASCII16 },	/* Zanac-EX (1986) (Pony Cannon) (J) */
	{ 0x2605031E,	CARTTYPE_ASCII16 },	/* Zoids (1988) (Toshiba Emi) (J) */
	
	{ 0xC13004C5,	CARTTYPE_ASCII16 },	/* Zemmix 126-in-1 (K) (hacked?) */
	
	/* ASCII 16 + SRAM */
	{ 0x96B7FACA,	CARTTYPE_ASCII16S },	/* Harry Fox MSX Special (1986) (MicroCabin) (J) */
	{ 0xD8055F5F,	CARTTYPE_ASCII16S },	/* Hydlide 2 (1986) (T&E Soft) (J) */
	
	{ 0xC956E1DF,	CARTTYPE_ASCII16S },	/* A-Train (1989) (Pony Cannon) (J) */
	{ 0xA987A0D8,	CARTTYPE_ASCII16S },	/* Daisenryaku (1986) (Micro Cabin) (J) */
	{ 0x00B8E82E,	CARTTYPE_ASCII16S },	/* Jyansei (1986) (Sony) (J) */
	{ 0x2C2924ED,	CARTTYPE_ASCII16S },	/* Jyansei (1986) (Sony) (J) [a1] */
	{ 0x2ABB1EC4,	CARTTYPE_ASCII16S },	/* Professional Mahjong Gokuh (1988) (Chatnoir) (J) */
	{ 0x8DF9945A,	CARTTYPE_ASCII16S },	/* Super Daisenryaku (1988) (Micro Cabin) (J) */
	
	
	/* Zemina types */
	{ 0x10D97B9E,	CARTTYPE_ZEMINAZMB },	/* 1942 */
	{ 0x8BD9D286,	CARTTYPE_ZEMINAZMB },	/* 1942 (alt) */
	{ 0x4F70C1F9,	CARTTYPE_ZEMINAZMB },	/* 3 Eyed One */
	{ 0xC73D4D25,	CARTTYPE_ZEMINAZMB },	/* Fantasy Zone */
	{ 0x58DE803B,	CARTTYPE_ZEMINAZMB },	/* Hydlide 2 (doesn't work, probably wrong emulation) */
	{ 0x41C82156,	CARTTYPE_ZEMINAZMB },	/* Hydlide 3 */
	{ 0x8E0F0638,   CARTTYPE_ZEMINAZMB },	/* Legendly Knight */
	{ 0x83F0EEDE,	CARTTYPE_KONAMI },	/* Street Master (cartridge, could as well be ZMB mapper specs with ROM though) */
	{ 0x112ABF34,	CARTTYPE_KONAMI },	/* Super Bio Man 4 (see Street Master note) */
	{ 0x32F11DF4,	CARTTYPE_ZEMINAZMB },	/* Super Laydock */
	{ 0xB885A464,	CARTTYPE_ZEMINAZMB },	/* Super Laydock (alt) */
	{ 0x87361B76,	CARTTYPE_ZEMINAZMB },	/* Valis */
	
	/* Bootleg 80-in-1 */
	{ 0x8D452FA1,	CARTTYPE_BTL80 },	/* Zemmix 30-in-1 (K) */
	{ 0xE50BAB84,	CARTTYPE_BTL80 },	/* Zemmix 64-in-1 (K) */
	{ 0xD9AA0055,	CARTTYPE_BTL80 },	/* Zemmix 80-in-1 (K) */
	
	/* Bootleg 90-in-1 */
	{ 0x123BE4E7,	CARTTYPE_BTL90 },	/* Zemmix 90-in-1 (K) */
	
	/* Bootleg 126-in-1 */
	{ 0x2EE35C23,	CARTTYPE_BTL126 },	/* Zemmix 126-in-1 (K) */
	
	
	/* dB-Soft Cross Blaim */
	{ 0x47273220,	CARTTYPE_CROSSBLAIM },	/* Cross Blaim (1985) (dB-Soft) (J) */
	
	/* MicroCabin Harry Fox */
	{ 0xC7FE3EE1,	CARTTYPE_HARRYFOX },	/* Harry Fox - Yuki no Maou hen (1985) (MicroCabin) (J) */
	
	
	/* Konami */
	{ 0x4D44255F,	CARTTYPE_KONAMI },	/* Gradius (1986) (Konami) (J) */
	{ 0x4DFCC009,	CARTTYPE_KONAMI },	/* Gradius (1986) (Konami) (J) [a1] */
	{ 0xFE23D253,	CARTTYPE_KONAMI },	/* Maze of Galious, The (1987) (Konami) (J) */
	{ 0xCF60FA7D,	CARTTYPE_KONAMI },	/* Shalom (1987) (Konami) (J) */
	{ 0x0F6418D3,	CARTTYPE_KONAMI },	/* Yume Tairiku Adventure (1986) (Konami) (J) (Penguin Adventure) */
	
	{ 0x5953D084,	CARTTYPE_KONAMI },	/* Akumajyo Drakyula - Vampire Killer (1986) (Konami) (J) */
	{ 0x0AEDEA40,	CARTTYPE_KONAMI },	/* Akumajyo Drakyula - Vampire Killer (1986) (Konami) (J) [a1] */
	{ 0x038C250D,	CARTTYPE_KONAMI },	/* Akumajyo Drakyula - Vampire Killer (1986) (Konami) (J) [a2] */
	{ 0x217935D0,	CARTTYPE_KONAMI },	/* Akumajyo Drakyula - Vampire Killer (1986) (Konami) (J) [a3] */
	{ 0x40E53EA3,	CARTTYPE_KONAMI },	/* Ganbare Goemon (1987) (Konami) (J) */
	{ 0xD95116B0,	CARTTYPE_KONAMI },	/* Hinotori (1987) (Konami) (J) */
	{ 0x3AB5AE57,	CARTTYPE_KONAMI },	/* Hinotori (1987) (Konami) (J) [a1] */
	{ 0x23B78A2D,	CARTTYPE_KONAMI },	/* King Kong 2 (1986) (Konami) (J) */
	{ 0xFAFE1303,	CARTTYPE_KONAMI },	/* Metal Gear 1 (1987) (Konami) (J) */
	{ 0x30396650,	CARTTYPE_KONAMI },	/* Usas (1987) (Konami) (J) */
	{ 0x1F1F1F26,	CARTTYPE_KONAMI },	/* Usas (1987) (Konami) (J) [a1] */
	
	{ 0x92016876,	CARTTYPE_KONAMI },	/* Gradius (Konami Antiques Collection 1) */
	{ 0xB10D32D8,	CARTTYPE_KONAMI },	/* Gradius (Bonus) (Konami Antiques Collection 1) */
	{ 0xF9E5DAF9,	CARTTYPE_KONAMI },	/* Yume Tairiku Adventure (Konami Antiques Collection 3) */
	
	/* Konami Game Master 2 */
	{ 0x084C5803,	CARTTYPE_GAMEMASTER2 },	/* Konami's Game Master 2 (1987) (Konami) (J) */
	
	/* Konami Synthesizer */
	{ 0xB9B0999A,	CARTTYPE_KONAMISYN },	/* Konami's Synthesizer (1986) (Konami) (J) */
	
	
	/* Konami SCC */
	{ 0x2608C959,	CARTTYPE_KONAMISCC },	/* A1 Spirit (1987) (Konami) (J) */
	{ 0x12BFD3A9,	CARTTYPE_KONAMISCC },	/* F1 Spirit (1987) (Konami) (J) */
	{ 0x64D2DF7C,	CARTTYPE_KONAMISCC },	/* F1 Spirit (1987) (Konami) (J) [a1] */
	{ 0x4B61AE91,	CARTTYPE_KONAMISCC },	/* Gofer no Yabou Episode 2 (1988) (Konami) (J) */
	{ 0x0DB7132F,	CARTTYPE_KONAMISCC },	/* Gofer no Yabou Episode 2 (1988) (Konami) (J) [a1] */
	{ 0x329987BC,	CARTTYPE_KONAMISCC },	/* Gofer no Yabou Episode 2 (Korean) */
	{ 0xDB847B2B,	CARTTYPE_KONAMISCC },	/* Gradius 2 (1987) (Konami) (J) */
	{ 0x32ABA4E3,	CARTTYPE_KONAMISCC },	/* Gradius 2 (1987) (Konami) (J) [a1] */
	{ 0xC8FAC21A,	CARTTYPE_KONAMISCC },	/* Gradius 2 (Demo Version) (1987) (Konami) (J) */
	{ 0xDFA8C827,	CARTTYPE_KONAMISCC },	/* Gradius 2 (Demo Version) (1987) (Konami) (J) (alt) */
	{ 0xE82A8E8E,	CARTTYPE_KONAMISCC },	/* King's Valley 2 (1988) (Konami) (J) */
	{ 0x9BB308F5,	CARTTYPE_KONAMISCC },	/* Parodius (1988) (Konami) (J) */
	{ 0xC36F559D,	CARTTYPE_KONAMISCC },	/* Salamander (1987) (Konami) (J) */
	
	{ 0xC08FAF0E,	CARTTYPE_KONAMISCC },	/* Gekitotsu Pennant Race (1988) (Konami) (J) */
	{ 0xB0A7800F,	CARTTYPE_KONAMISCC },	/* Gekitotsu Pennant Race (1988) (Konami) (J) [a1] */
	{ 0xC44C6330,	CARTTYPE_KONAMISCC },	/* Gekitotsu Pennant Race 2 (1989) (Konami) (J) */
	{ 0xB1CB1052,	CARTTYPE_KONAMISCC },	/* King's Valley 2 (1988) (Konami) (J) */
	{ 0x774C70FC,	CARTTYPE_KONAMISCC },	/* King's Valley 2 (Edit Contest Version) (1988) (Konami) (J) */
	{ 0x4E82660D,	CARTTYPE_KONAMISCC },	/* Kontora - Contra (1989) (Konami) (J) */
	{ 0x18B1F34B,	CARTTYPE_KONAMISCC },	/* Metal Gear 2 - Solid Snake (1990) (Konami) (J) */
	{ 0xDCEE5080,	CARTTYPE_KONAMISCC },	/* Metal Gear 2 - Solid Snake (Demo Version) (1990) (Konami) (J) */
	{ 0xA0014E7F,	CARTTYPE_KONAMISCC },	/* Quarth (1990) (Konami) (J) */
	{ 0xC4CCB75F,	CARTTYPE_KONAMISCC },	/* Quarth (Demo Version) (1990) (Konami) (J) */
	{ 0x27405EED,	CARTTYPE_KONAMISCC },	/* Space Manbow (1989) (Konami) (J) */
	
	{ 0x7C216BDE,	CARTTYPE_KONAMISCC },	/* Gofer no Yabou Episode 2 (Konami Antiques Collection 1) */
	{ 0xC0E331DA,	CARTTYPE_KONAMISCC },	/* Gradius 2 (Konami Antiques Collection 2) */
	
	{ 0xBFBB4993,	CARTTYPE_KONAMISCC },	/* dvik&joyrex's Aliens demo */
	{ 0xB0638377,	CARTTYPE_KONAMISCC },	/* dvik&joyrex's Leila K demo */
	{ 0x6795EC3C,	CARTTYPE_KONAMISCC },	/* Infinite's Montana John and the Templar's Treasure */
	{ 0x7B5E126F,	CARTTYPE_KONAMISCC },	/* Infinite's Montana John and the Templar's Treasure v1.1 */
	{ 0xB88A7FDE,	CARTTYPE_KONAMISCC },	/* Jos'b's Los Jardines de Zee Wang Zu (SCC version) */
	{ 0xD6C395F8,	CARTTYPE_KONAMISCC },	/* Nerlaska's Monster Hunter (English) */
	{ 0xB876A92F,	CARTTYPE_KONAMISCC },	/* Nerlaska's Monster Hunter (Spanish) */
	{ 0xAA6F2968,	CARTTYPE_KONAMISCC },	/* Nerlaska's Mr. Mole */
	
	/* Konami Game Collection SCC hacks */
	{ 0xF3E0F940,	CARTTYPE_KONAMISCC },	/* Antarctic Adventure */
	{ 0x04391BC8,	CARTTYPE_KONAMISCC },	/* Hyper Sports 2 */
	{ 0x9AB41FF5,	CARTTYPE_KONAMISCC },	/* Knightmare */
	{ 0xE4A8EE40,	CARTTYPE_KONAMISCC },	/* Konami's Boxing */
	{ 0x37183EFD,	CARTTYPE_KONAMISCC },	/* Super Cobra */
	{ 0xB29F137C,	CARTTYPE_KONAMISCC },	/* Video Hustler */
	{ 0xE36710CE,	CARTTYPE_KONAMISCC },	/* Yie ar Kung-Fu */
	{ 0x4CF4D4CF,	CARTTYPE_KONAMISCC },	/* Yie ar Kung-Fu 2 */
	
	/* Konami Game Collection The Snatcher SCC-I hacks */
	{ 0x037D1ED8,	CARTTYPE_KONAMISCCI },	/* Antarctic Adventure */
	{ 0x2CA8E6BF,	CARTTYPE_KONAMISCCI },	/* Hyper Rally */
	{ 0x62F1531C,	CARTTYPE_KONAMISCCI },	/* Hyper Sports 2 */
	{ 0xDBC8B11B,	CARTTYPE_KONAMISCCI },	/* Knightmare */
	{ 0x0E8EFD8B,	CARTTYPE_KONAMISCCI },	/* Konami's Boxing */
	{ 0x619B4B4A,	CARTTYPE_KONAMISCCI },	/* Nemesis */
	{ 0x9EA714C2,	CARTTYPE_KONAMISCCI },	/* Pippols */
	{ 0x39B45611,	CARTTYPE_KONAMISCCI },	/* Road Fighter */
	{ 0x1B47F46E,	CARTTYPE_KONAMISCCI },	/* Sky Jaguar */
	{ 0x8A4B51BC,	CARTTYPE_KONAMISCCI },	/* Super Cobra */
	{ 0x4A653152,	CARTTYPE_KONAMISCCI },	/* Twinbee */
	{ 0xF338052F,	CARTTYPE_KONAMISCCI },	/* Video Hustler */
	{ 0x278114CD,	CARTTYPE_KONAMISCCI },	/* Yie ar Kung-Fu */
	{ 0xC080159C,	CARTTYPE_KONAMISCCI },	/* Yie ar Kung-Fu 2 */
	
	/* Konami Game Collection SD Snatcher SCC-I hacks */
	{ 0x14CCB2FD,	CARTTYPE_KONAMISCCI },	/* Hyper Rally */
	{ 0x04A889BF,	CARTTYPE_KONAMISCCI },	/* Hyper Sports 2 */
	{ 0x408070C4,	CARTTYPE_KONAMISCCI },	/* Konami's Boxing */
	{ 0x34C903BB,	CARTTYPE_KONAMISCCI },	/* Nemesis (doesn't work) */
	{ 0x1EC24C26,	CARTTYPE_KONAMISCCI },	/* Pippols */
	{ 0x6616395A,	CARTTYPE_KONAMISCCI },	/* Road Fighter */
	{ 0x97A62DBA,	CARTTYPE_KONAMISCCI },	/* Twinbee */
	{ 0x7A5FEC5A,	CARTTYPE_KONAMISCCI },	/* Video Hustler */
	
	/* Manuel Mega Flash ROM SCC */
	{ 0x3B38CFEA,	CARTTYPE_MFLASHSCC },	/* dvik&joyrex's Lotus F3 */
	
	/* Matra INK */
	{ 0x35823FD5,	CARTTYPE_MATRAINK },	/* INK - Exxon Surfing (2006) (Matra) */
	
	
	/* Sony Playball */
	{ 0xD178833B,	CARTTYPE_PLAYBALL },	/* Playball (1986) (Sony) (J) */
	
	
	/* end */
	{ 0xFFFFFFFF,	CARTTYPE_MAX }	
};
