#include <windows.h>
#include <shlobj.h>

#include "global.h"
#include "settings.h"
#include "resource.h"
#include "main.h"
#include "file.h"
#include "media.h"
#include "mapper.h"
#include "msx.h"
#include "netplay.h"
#include "cont.h"
#include "draw.h"
#include "crystal.h"
#include "movie.h"
#include "psg.h"
#include "psglogger.h"
#include "tape.h"
#include "reverse.h"
#include "vdp.h"

enum {
	MEDIA_BIOS=0,
	MEDIA_CART1,
	MEDIA_CART2,
	MEDIA_TAPE,
	
	MEDIA_MAX
};

enum {
	MEDIA_TAB_CARTS=0,
	MEDIA_TAB_TAPE,
	MEDIA_TAB_INTERNAL,
	
	MEDIA_TAB_MAX
};

static struct {
	int fi_bios;
	int fi_cart_set;
	int fi_cart1;
	int fi_cart2;
	int fi_tape;
	
	int current_tab;
	int prev_tab;
	
	char last_of[STRING_SIZE];
	int shift_open;
	int shift_cart[2];
	
	u32 tape_crc_prev;
	int tape_insertion_changed;
} media;

int media_get_filterindex_bios(void) { return media.fi_bios; }
int media_get_filterindex_tape(void) { return media.fi_tape; }
int media_get_filterindex_cart(void) { return media.fi_cart_set; }

void media_init(void)
{
	memset(&media,0,sizeof(media));
	
	/* file filter index settings */
	media.fi_bios=1;	SETTINGS_GET_INT(settings_info(SETTINGS_FILTERINDEX_BIOS),&media.fi_bios);	CLAMP(media.fi_bios,1,2);
	media.fi_tape=1;	SETTINGS_GET_INT(settings_info(SETTINGS_FILTERINDEX_TAPE),&media.fi_tape);	CLAMP(media.fi_tape,1,2);
	media.fi_cart_set=1;	SETTINGS_GET_INT(settings_info(SETTINGS_FILTERINDEX_CART),&media.fi_cart_set);	CLAMP(media.fi_cart_set,1,3);
	media.fi_cart1=media.fi_cart2=media.fi_cart_set;
}



static void media_log_type(int slot,int one)
{
	char ls[0x10];
	if (mapper_get_cartsize(slot)==0) sprintf(ls,"empty");
	else {
		int size=mapper_get_cartsize(slot)>>10;
		if (size==0) size=1;
		sprintf(ls,"%dK",size);
	}
	
	LOG(LOG_MISC,"%s%d: %s %s",one?"s":"slot ",slot+1,ls,mapper_get_type_shortname(mapper_get_carttype(slot)));
}

int media_open_single(const char* fn)
{
	/* try cas */
	file_setfile(NULL,fn,NULL,"cas");
	if (file->ext&&(strlen(file->ext)==3)&&(stricmp(file->ext,"cas")==0)) {
		tape_cur_to_od();
		tape_flush_cache_od();
		if (!tape_save_od()) LOG(LOG_MISC|LOG_WARNING,"couldn't save tape\n");
		
		if (tape_open_od(TRUE)) {
			char c[STRING_SIZE];
			
			media.prev_tab=MEDIA_TAB_TAPE;
			
			reverse_invalidate();
			tape_od_to_cur();
			tape_set_runtime_block_inserted(FALSE);
			tape_log_info();
			mapper_update_bios_hack();
			
			main_titlebar(mapper_get_current_name(c));
			
			return 2;
		}
		
		tape_clean_od();
	}
	
	/* try rom */
	file_setfile(NULL,fn,NULL,"rom");
	if (mapper_open_cart(0,fn,TRUE)) {
		char c[STRING_SIZE];
		
		media.prev_tab=MEDIA_TAB_CARTS;
		
		mapper_set_carttype(0,mapper_autodetect_cart(fn));
		
		main_titlebar(mapper_get_current_name(c));
		
		media_log_type(0,FALSE);
		LOG(LOG_MISC,"\n");
		
		return 1;
	}
	
	return 0;
}

void media_drop(WPARAM p)
{
	char fn[STRING_SIZE]={0};
	HDROP drop=(HDROP)p;
	UINT numfiles=DragQueryFile(drop,~0,NULL,0);
	
	if (netplay_is_active()) { LOG(LOG_MISC|LOG_WARNING,"can't open files during netplay\n"); return; }
	if (movie_get_active_state()) { LOG(LOG_MISC|LOG_WARNING,"can't open files during movie\n"); return; }
	if (psglogger_is_active()) { LOG(LOG_MISC|LOG_WARNING,"can't open files during PSG log\n"); return; }
	
	DragQueryFile(drop,0,fn,STRING_SIZE);
	DragFinish(drop);
	
	if (numfiles!=1) { LOG(LOG_MISC|LOG_WARNING,"can't open multiple files\n"); return; }
	
	msx_wait();
	if (media_open_single(fn)) {
		mapper_init_cart(0);
		msx_reset(TRUE);
	}
	msx_wait_done();
	
	SetForegroundWindow(MAIN->window);
}




static int openfilenamedialog(HWND dialog,UINT type)
{
	const char* filter[]={
		"BIOS ROMs (*.rom, *.zip)\0*.rom;*.ri;*.zip\0All Files\0*.*\0\0",
		"All Supported Media\0*.rom;*.ri;*.dsk;*.di1;*.di2;*.zip\0MSX ROMs (*.rom, *.zip)\0*.rom;*.ri;*.zip\0MSX Disks (*.dsk, *.zip)\0*.dsk;*.di1;*.di2;*.zip\0All Files\0*.*\0\0",
		"fMSX CAS Files (*.cas, *.zip)\0*.cas;*.zip\0All Files\0*.*\0\0"
	};
	const char* title[]={
		"Open BIOS",
		"Open Cartridge 1",
		"Open Cartridge 2",
		"Open Tape"
	};
	char fn[STRING_SIZE]={0};
	OPENFILENAME of;
	BOOL o;
	
	/* type-specific */
	UINT eid=IDC_MEDIA_SLOT1FILE;	/* editbox id */
	char* dir=file->romdir;		/* initial dir if file is invalid */
	int* fi=&media.fi_cart1;	/* filterindex */
	
	media.shift_open=GetAsyncKeyState(VK_SHIFT);
	
	memset(&of,0,sizeof(OPENFILENAME));
	
	switch (type) {
		case MEDIA_BIOS: eid=IDC_MEDIA_BIOSFILE; dir=file->biosdir; fi=&media.fi_bios; of.lpstrFilter=filter[0]; of.lpstrTitle=title[0]; break;
		case MEDIA_CART1: eid=IDC_MEDIA_SLOT1FILE; dir=file->romdir; fi=&media.fi_cart1; of.lpstrFilter=filter[1]; of.lpstrTitle=title[1]; break;
		case MEDIA_CART2: eid=IDC_MEDIA_SLOT2FILE; dir=file->romdir; fi=&media.fi_cart2; of.lpstrFilter=filter[1]; of.lpstrTitle=title[2]; break;
		case MEDIA_TAPE: eid=IDC_MEDIA_TAPEFILE; dir=file->casdir; fi=&media.fi_tape; of.lpstrFilter=filter[2]; of.lpstrTitle=title[3]; break;
		
		default: break;
	}
	
	/* check+set initial dir/filename */
	if (GetDlgItemText(dialog,eid,fn,STRING_SIZE)) {
		file_setfile(NULL,fn,NULL,type==MEDIA_TAPE?"cas":"rom");
		if (file->filename&&file_open()) strcpy(fn,file->filename);
		else memset(fn,0,STRING_SIZE);
		file_close();
	}
	
	/* opendialog */
	of.lStructSize=sizeof(OPENFILENAME);
	of.hwndOwner=dialog;
	of.hInstance=MAIN->module;
	of.lpstrFile=fn;
	of.nMaxFile=STRING_SIZE;
	of.nFilterIndex=*fi;
	of.Flags=OFN_ENABLESIZING|OFN_HIDEREADONLY|OFN_FILEMUSTEXIST;
	of.lpstrInitialDir=strlen(fn)?NULL:dir?dir:file->appdir;
	
	o=GetOpenFileName(&of);
	*fi=of.nFilterIndex;
	
	media.shift_open=(media.shift_open|GetAsyncKeyState(VK_SHIFT))&0x8000;
	
	if (!o) return FALSE;
	
	media.last_of[0]=0;
	if (strlen(fn)) strcpy(media.last_of,fn);
	
	return TRUE;
}


static void media_tab(HWND dialog,int tab)
{
	int i;
	
	if (tab==media.current_tab) return;
	
	media.current_tab=tab;
	
	#define HIDETAB(x,y) for (i=x;i<=y;i++) { EnableWindow(GetDlgItem(dialog,i),FALSE); ShowWindow(GetDlgItem(dialog,i),SW_HIDE); }
	#define SHOWTAB(x,y) for (i=x;i<=y;i++) { EnableWindow(GetDlgItem(dialog,i),TRUE); ShowWindow(GetDlgItem(dialog,i),SW_NORMAL); }
	
	/* hide+disable all */
	HIDETAB(IDC_MEDIA_TAB0START,IDC_MEDIA_TAB0END)
	HIDETAB(IDC_MEDIA_TAB1START,IDC_MEDIA_TAB1END)
	HIDETAB(IDC_MEDIA_TAB2START,IDC_MEDIA_TAB2END)
	
	switch (tab) {
		/* carts */
		case MEDIA_TAB_CARTS:
			SHOWTAB(IDC_MEDIA_TAB0START,IDC_MEDIA_TAB0END)
			break;
		
		/* tape */
		case MEDIA_TAB_TAPE:
			SHOWTAB(IDC_MEDIA_TAB1START,IDC_MEDIA_TAB1END)
			EnableWindow(GetDlgItem(dialog,IDC_MEDIA_TAPER),FALSE);
			EnableWindow(GetDlgItem(dialog,IDC_MEDIA_TAPEPOS),FALSE);
			EnableWindow(GetDlgItem(dialog,IDC_MEDIA_TAPESIZE),FALSE);
			
			tape_draw_gui(dialog);
			
			PostMessage(dialog,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(dialog,IDC_MEDIA_TAPECONTENTS),TRUE);
			break;
		
		/* internal */
		case MEDIA_TAB_INTERNAL:
			SHOWTAB(IDC_MEDIA_TAB2START,IDC_MEDIA_TAB2END)
			break;
		
		default: break;
	}
	
	#undef HIDETAB
	#undef SHOWTAB
}


BOOL CALLBACK media_dialog(HWND dialog,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch (msg) {
		case WM_INITDIALOG: {
			int i;
			HWND c1=GetDlgItem(dialog,IDC_MEDIA_SLOT1TYPE);
			HWND c2=GetDlgItem(dialog,IDC_MEDIA_SLOT2TYPE);
			
			TCITEM pitem;
			HWND tab=GetDlgItem(dialog,IDC_MEDIA_TAB);
			const char* tabtext[]={"Cartridges","Tape","Internal"};
			
			memset(&pitem,0,sizeof(TCITEM));
			
			/* init rom editboxes */
			if (mapper_get_file(0)) SetDlgItemText(dialog,IDC_MEDIA_SLOT1FILE,mapper_get_file(0));
			if (mapper_get_file(1)) SetDlgItemText(dialog,IDC_MEDIA_SLOT2FILE,mapper_get_file(1));
			
			/* init rom comboboxes */
			for (i=0;i<CARTTYPE_MAX;i++) {
				SendMessage(c1,CB_ADDSTRING,0,(LPARAM)((LPCTSTR)mapper_get_type_longname(i)));
				SendMessage(c2,CB_ADDSTRING,0,(LPARAM)((LPCTSTR)mapper_get_type_longname(i)));
			}
			SendMessage(c1,CB_SETCURSEL,mapper_get_carttype(0),0);
			SendMessage(c2,CB_SETCURSEL,mapper_get_carttype(1),0);
			media.shift_cart[0]=media.shift_cart[1]=0;
			
			/* internal comboboxes */
			c1=GetDlgItem(dialog,IDC_MEDIA_REGION);
			for (i=0;i<CONT_REGION_MAX;i++) SendMessage(c1,CB_ADDSTRING,0,(LPARAM)((LPCTSTR)cont_get_region_name(i)));
			SendMessage(c1,CB_SETCURSEL,cont_get_region(),0);
			
			c1=GetDlgItem(dialog,IDC_MEDIA_VDP);
			for (i=0;i<VDP_CHIPTYPE_MAX;i++) SendMessage(c1,CB_ADDSTRING,0,(LPARAM)((LPCTSTR)vdp_get_chiptype_name(i)));
			SendMessage(c1,CB_SETCURSEL,vdp_get_chiptype(),0);
			
			c1=GetDlgItem(dialog,IDC_MEDIA_PSG);
			for (i=0;i<PSG_CHIPTYPE_MAX;i++) SendMessage(c1,CB_ADDSTRING,0,(LPARAM)((LPCTSTR)psg_get_chiptype_name(i)));
			SendMessage(c1,CB_SETCURSEL,psg_get_chiptype(),0);
			
			if (mapper_get_bios_file()) SetDlgItemText(dialog,IDC_MEDIA_BIOSFILE,mapper_get_bios_file());
			
			/* init tape */
			media.tape_insertion_changed=FALSE;
			msx_wait(); tape_cur_to_od(); msx_wait_done();
			tape_flush_cache_od();
			media.tape_crc_prev=tape_get_crc_od();
			tape_init_gui(SetWindowLongPtr(GetDlgItem(dialog,IDC_MEDIA_TAPECONTENTS),GWLP_WNDPROC,(LONG_PTR)tape_sub_listbox));
			tape_draw_listbox(GetDlgItem(dialog,IDC_MEDIA_TAPECONTENTS));
			tape_draw_gui(dialog);
			
			/* init tab */
			media.current_tab=-1;
			
			pitem.mask=TCIF_TEXT;
			pitem.iImage=-1;
			
			pitem.pszText=(LPTSTR)tabtext[0];	SendMessage(tab,TCM_INSERTITEM,0,(LPARAM)&pitem);
			pitem.pszText=(LPTSTR)tabtext[1];	SendMessage(tab,TCM_INSERTITEM,1,(LPARAM)&pitem);
			pitem.pszText=(LPTSTR)tabtext[2];	SendMessage(tab,TCM_INSERTITEM,2,(LPARAM)&pitem);
			
			SendMessage(tab,TCM_SETCURSEL,media.prev_tab,0);
			media_tab(dialog,media.prev_tab);
			
			
			main_parent_window(dialog,0,0,0,0,0);
			
			return 1;
			break;
		}
		
		case WM_CTLCOLORSTATIC: /* case WM_CTLCOLORBTN: */ {
			UINT id=GetDlgCtrlID((HWND)lParam);
			if (id==IDC_MEDIA_TAPEFILE) break;
			
			SetBkMode((HDC)wParam,TRANSPARENT);
			return (BOOL)(HBRUSH)GetStockObject(NULL_BRUSH);
		}
		
		case WM_NOTIFY:
			switch (LOWORD(wParam)) {
				
				/* tab changes */
				case IDC_MEDIA_TAB:
					if (((LPOFNOTIFY)lParam)->hdr.code==TCN_SELCHANGE) {
						int tab;
						if ((tab=SendDlgItemMessage(dialog,IDC_MEDIA_TAB,TCM_GETCURSEL,0,0))!=-1) media_tab(dialog,tab);
					}
					break;
				
				default: break;
			}
			
			break;
		
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				
				/* bios */
				case IDC_MEDIA_BIOSBROWSE:
					if (openfilenamedialog(dialog,MEDIA_BIOS)) SetDlgItemText(dialog,IDC_MEDIA_BIOSFILE,media.last_of);
					PostMessage(dialog,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(dialog,IDOK),TRUE);
					break;
				case IDC_MEDIA_BIOSDEFAULT:
					SetDlgItemText(dialog,IDC_MEDIA_BIOSFILE,mapper_get_defaultbios_file());
					PostMessage(dialog,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(dialog,IDOK),TRUE);
					break;
				
				/* cart file browse buttons */
				#define OPENDC(slot) \
					if (openfilenamedialog(dialog,MEDIA_CART1+(slot-1))) { \
						media.shift_cart[slot-1]=media.shift_open; \
						SetDlgItemText(dialog,IDC_MEDIA_SLOT1FILE+(slot-1),media.last_of); \
						SendDlgItemMessage(dialog,IDC_MEDIA_SLOT1TYPE+(slot-1),CB_SETCURSEL,mapper_autodetect_cart(media.last_of),0); \
						file_setfile(NULL,media.last_of,NULL,"rom"); file_setdirectory(&file->romdir); /* remember dir */ \
					} \
					media.fi_cart_set=media.fi_cart##slot; \
					PostMessage(dialog,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(dialog,IDOK),TRUE)
				
				case IDC_MEDIA_SLOT1BROWSE: OPENDC(1); break;
				case IDC_MEDIA_SLOT2BROWSE: OPENDC(2); break;
				
				/* cart eject buttons */
				case IDC_MEDIA_SLOT1EJECT:
					SetDlgItemText(dialog,IDC_MEDIA_SLOT1FILE,"");
					SendDlgItemMessage(dialog,IDC_MEDIA_SLOT1TYPE,CB_SETCURSEL,CARTTYPE_MIRROR,0);
					PostMessage(dialog,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(dialog,IDOK),TRUE);
					break;
				case IDC_MEDIA_SLOT2EJECT:
					SetDlgItemText(dialog,IDC_MEDIA_SLOT2FILE,"");
					SendDlgItemMessage(dialog,IDC_MEDIA_SLOT2TYPE,CB_SETCURSEL,CARTTYPE_MIRROR,0);
					PostMessage(dialog,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(dialog,IDOK),TRUE);
					break;
				
				/* tape browse */
				case IDC_MEDIA_TAPEBROWSE:
					/* openfilenamedialog here instead of tape.c, tapes can be opened elsewhere too */
					if (openfilenamedialog(dialog,MEDIA_TAPE)&&strlen(media.last_of)) {
						if (!tape_save_od()) LOG_ERROR_WINDOW(dialog,"Couldn't save tape!");
						file_setfile(NULL,media.last_of,NULL,"cas");
						if (tape_open_od(media.shift_open==0)) {
							file_setdirectory(&file->casdir); /* remember dir */
							tape_draw_listbox(GetDlgItem(dialog,IDC_MEDIA_TAPECONTENTS));
							tape_draw_gui(dialog);
							tape_set_updated_od();
							media.tape_insertion_changed=TRUE;
						}
						else {
							LOG_ERROR_WINDOW(dialog,"Couldn't open tape!");
						}
					}
					PostMessage(dialog,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(dialog,IDC_MEDIA_TAPECONTENTS),TRUE);
					InvalidateRect(dialog,NULL,FALSE);
					break;
				
				/* tape eject */
				case IDC_MEDIA_TAPEEJECT:
					if (!tape_save_od()) LOG_ERROR_WINDOW(dialog,"Couldn't save tape!");
					tape_clean_od();
					tape_draw_listbox(GetDlgItem(dialog,IDC_MEDIA_TAPECONTENTS));
					tape_draw_gui(dialog);
					tape_set_updated_od();
					media.tape_insertion_changed=TRUE;
					PostMessage(dialog,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(dialog,IDOK),TRUE);
					break;
				
				/* tape new */
				case IDC_MEDIA_TAPENEW:
					if (tape_new_od(dialog)) {
						tape_draw_listbox(GetDlgItem(dialog,IDC_MEDIA_TAPECONTENTS));
						tape_draw_gui(dialog);
						tape_set_updated_od();
						media.tape_insertion_changed=TRUE;
					}
					PostMessage(dialog,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(dialog,IDC_MEDIA_TAPECONTENTS),TRUE);
					break;
				
				/* tape listbox */
				case IDC_MEDIA_TAPECONTENTS:
					switch (HIWORD(wParam)) {
						
						/* clicked/changed */
						case LBN_SELCHANGE: {
							LRESULT cb=SendDlgItemMessage(dialog,IDC_MEDIA_TAPECONTENTS,LB_GETCURSEL,0,0);
							if (cb==LB_ERR) break;;
							
							tape_set_cur_block_od(cb);
							tape_draw_pos(dialog);
							tape_set_updated_od();
							
							break;
						}
						
						/* doubleclick = ok+reset */
						case LBN_DBLCLK:
							if (!GetListBoxInfo(GetDlgItem(dialog,IDC_MEDIA_TAPECONTENTS))) break;
							PostMessage(dialog,WM_COMMAND,IDOK,0);
							break;
						
						default: break;
					}
					
					break;
				
				/* ok/ok + reset button */
				case IDOK: case IDC_MEDIA_OK: {
					char fn[STRING_SIZE]={0};
					int c,reset=(LOWORD(wParam)==IDOK);
					int ci[2]={FALSE,FALSE};
					
					if (media.current_tab!=-1) media.prev_tab=media.current_tab;
					
					msx_wait();
					
					/* internal tab */
					/* bios */
					if (GetDlgItemText(dialog,IDC_MEDIA_BIOSFILE,fn,STRING_SIZE)==0) strcpy(fn,mapper_get_defaultbios_file());
					mapper_open_bios(fn);
					
					/* region */
					if ((c=SendDlgItemMessage(dialog,IDC_MEDIA_REGION,CB_GETCURSEL,0,0))!=CB_ERR) cont_set_region(c);
					
					/* vdp */
					if ((c=SendDlgItemMessage(dialog,IDC_MEDIA_VDP,CB_GETCURSEL,0,0))!=CB_ERR) {
						vdp_set_chiptype(c);
						draw_set_vidformat(vdp_get_chiptype_vf());
						crystal_set_mode(reset^1);
					}
					
					/* psg */
					if ((c=SendDlgItemMessage(dialog,IDC_MEDIA_PSG,CB_GETCURSEL,0,0))!=CB_ERR) {
						psg_set_chiptype(c);
						psg_init_amp();
					}
					
					/* carts tab */
					/* cart */
					#define OPENSLOT(slot) \
						if ((c=SendDlgItemMessage(dialog,IDC_MEDIA_SLOT1TYPE+slot,CB_GETCURSEL,0,0))==CB_ERR) c=CARTTYPE_MIRROR; \
						if (~mapper_get_type_flags(c)&MCF_EMPTY&&GetDlgItemText(dialog,IDC_MEDIA_SLOT1FILE+slot,fn,STRING_SIZE)) { \
							if (mapper_get_file(slot)==NULL||stricmp(fn,mapper_get_file(slot))!=0) { \
								if (mapper_open_cart(slot,fn,media.shift_cart[slot]==0)) { \
									mapper_set_carttype(slot,c); \
									if (!reset) mapper_init_cart(slot); \
									ci[slot]=TRUE; \
								} \
								else c=mapper_get_carttype(slot); \
							} \
						} \
						else if (mapper_get_file(slot)!=NULL) { \
							mapper_close_cart(slot); \
							mapper_set_carttype(slot,c); \
							if (!reset) mapper_init_cart(slot); \
							ci[slot]=TRUE; \
						} \
						if (c!=mapper_get_carttype(slot)) { \
							mapper_set_carttype(slot,c); \
							if (!reset) mapper_init_cart(slot); \
							ci[slot]=TRUE; \
						}
					
					OPENSLOT(0) OPENSLOT(1)
					
					/* show romtype */
					if (ci[0]) {
						media_log_type(0,ci[1]);
						if (ci[1]) LOG(LOG_MISC,", ");
						else LOG(LOG_MISC,"\n");
					}
					if (ci[1]) {
						media_log_type(1,ci[0]);
						LOG(LOG_MISC,"\n");
					}
					
					/* tape tab */
					c=SendDlgItemMessage(dialog,IDC_MEDIA_TAPECONTENTS,LB_GETCURSEL,0,0);
					if (c==LB_ERR) c=0;
					if (media.tape_crc_prev!=tape_get_crc_od()||tape_get_updated_od()) {
						/* changed */
						reverse_invalidate();
						tape_set_cur_block_od(c);
						tape_od_to_cur();
						tape_set_runtime_block_inserted(FALSE);
						
						if (media.tape_insertion_changed) tape_log_info();
					}
					else tape_clean_od();
					
					mapper_update_bios_hack();
					
					
					main_titlebar(mapper_get_current_name(fn));
					
					/* reset */
					if (reset) msx_reset(TRUE);
					else {
						mapper_refresh_pslot_read();
						mapper_refresh_pslot_write();
					}
					
					msx_wait_done();
					
					EndDialog(dialog,0);
					break;
				}
				
				case IDCANCEL:
					if (media.current_tab!=-1) media.prev_tab=media.current_tab;
					tape_clean_od(); /* clean temp tape */
					
					EndDialog(dialog,0);
					break;
				
				default: break;
			}
			break;
		
		default: break;
	}
	
	return 0;
}
